"""
utils module is part of DuIvyTools. This module provides log parent class.

Written by DuIvy and provided to you by GPLv3 license.
"""

import argparse
import logging
import sys
import time
from typing import List

from colorama import Back, Fore, Style


class log(object):
    """log class, a logging system parent class, provied five functions for output debug, info, warning, error, and critical messages."""

    logging.basicConfig(
        level=logging.INFO,
        ## TODO the line number
        ## format="%(asctime)s-%(filename)s[line:%(lineno)d]\n%(message)s",
        format="%(message)s",
    )
    logger = logging.getLogger(__name__)

    def debug(self, msg):
        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.logger.debug(
            Fore.CYAN + Back.WHITE + f"{time_info}\nDebug -> {msg}" + Style.RESET_ALL
        )

    def info(self, msg):
        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.logger.info(Fore.GREEN + f"{time_info}\nInfo -> {msg}" + Style.RESET_ALL)

    def warn(self, msg):
        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.logger.warning(
            Fore.YELLOW + f"{time_info}\nWarning -> {msg}" + Style.RESET_ALL
        )

    def error(self, msg):
        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.logger.error(
            Fore.WHITE + Back.RED + f"{time_info}\nError -> {msg}" + Style.RESET_ALL
        )
        sys.exit()

    def critical(self, msg):
        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.logger.critical(
            Fore.WHITE + Back.RED + f"{time_info}\nCRITICAL -> {msg}" + Style.RESET_ALL
        )
        sys.exit()


class Parameters(log):
    """A class to deal with and store user-input parameters"""

    def __init__(self) -> None:
        parser = argparse.ArgumentParser(
            description="DuIvyTools: A Simple MD Analysis Tool"
        )
        parser.add_argument("cmd", type=str, help="command of DIT to run")
        parser.add_argument(
            "-f", "--input", nargs="+", help="specify the input file or files"
        )
        parser.add_argument("-o", "--output", type=str, help="specify the output file")
        parser.add_argument(
            "-ns", "--noshow", action="store_true", help="not to show figure"
        )
        parser.add_argument(
            "-c",
            "--columns",
            nargs="+",
            help="select the column indexs for visualization or calculation, or input numerical list",
        )
        parser.add_argument(
            "-l", "--legends", nargs="+", help="specify the legends of figure or data"
        )
        parser.add_argument(
            "-b", "--begin", type=int, help="specify the index for beginning (include)"
        )
        parser.add_argument(
            "-e", "--end", type=int, help="specify the index for ending (not include)"
        )
        parser.add_argument(
            "-dt",
            "--dt",
            type=int,
            default=1,
            help="specify the index step, default to 1",
        )
        parser.add_argument(
            "-x", "--xlabel", type=str, help="specify the xlabel of figure or data"
        )
        parser.add_argument(
            "-y", "--ylabel", type=str, help="specify the ylabel of figure or data"
        )
        parser.add_argument(
            "-z", "--zlabel", type=str, help="specify the zlabel of figure or data"
        )
        parser.add_argument(
            "-t", "--title", type=str, help="specify the title of figure or data"
        )
        parser.add_argument(
            "-xmin", "--xmin", type=float, help="specify the X value limitation, x_min"
        )
        parser.add_argument(
            "-xmax", "--xmax", type=float, help="specify the X value limitation, x_max"
        )
        parser.add_argument(
            "-ymin", "--ymin", type=float, help="specify the Y value limitation, y_min"
        )
        parser.add_argument(
            "-ymax", "--ymax", type=float, help="specify the Y value limitation, y_max"
        )
        parser.add_argument(
            "-zmin", "--zmin", type=float, help="specify the Z value limitation, z_min"
        )
        parser.add_argument(
            "-zmax", "--zmax", type=float, help="specify the Z value limitation, z_max"
        )
        parser.add_argument(
            "--x_precision",
            type=int,
            help="specify the precision of X values for visualization",
        )
        parser.add_argument(
            "--y_precision",
            type=int,
            help="specify the precision of Y values for visualization",
        )
        parser.add_argument(
            "--z_precision",
            type=int,
            help="specify the precision of Z values for visualization",
        )
        parser.add_argument(
            "-xs",
            "--xshrink",
            type=float,
            default=1.0,
            help="modify X values by multipling xshrink, default to 1.0",
        )
        parser.add_argument(
            "-ys",
            "--yshrink",
            type=float,
            default=1.0,
            help="modify Y values by multipling yshrink, default to 1.0",
        )
        parser.add_argument(
            "-zs",
            "--zshrink",
            type=float,
            default=1.0,
            help="modify Z values by multipling zshrink, default to 1.0",
        )
        parser.add_argument(
            "-smv",
            "--showMV",
            action="store_true",
            help="whether to show moving averages of data",
        )
        parser.add_argument(
            "-ws",
            "--windowsize",
            type=int,
            default=50,
            help="window size for moving average calculation, default to 50",
        )
        parser.add_argument(
            "-cf",
            "--confidence",
            type=float,
            default=0.95,
            help="confidence for confidence interval calculation, default to 0.95",
        )
        parser.add_argument("--alpha", type=float, help="the alpha of figure items")
        parser.add_argument("-csv", "--csv", type=str, help="store data into csv file")
        parser.add_argument(
            "-eg",
            "--engine",
            type=str,
            default="matplotlib",
            choices=["matplotlib", "plotext", "plotly", "gnuplot"],
            help="specify the engine for plotting: 'matplotlib', 'plotext', 'plotly', 'gnuplot'",
        )
        parser.add_argument(
            "-cmap",
            "--colormap",
            help="specify the colormap applied for figures, available for 'matplotlib' and 'plotly' engine",
        )
        parser.add_argument(
            "--colorbar_location",
            type=str,
            default=None,
            choices=[None, "left", "top", "bottom", "right"],
            help="the location of colorbar, also determining the orientation of colorbar, ['left', 'top', 'bottom', 'right'], available for 'matplotlib'",
        )
        parser.add_argument(
            "--legend_location",
            type=str,
            default=None,
            choices=["inside", "outside"],
            help="the location of legend box, ['inside', 'outside'], available for 'matplotlib' and 'gnuplot'",
        )
        parser.add_argument(
            "-m",
            "--mode",
            type=str,
            choices=[
                "withoutScatter",
                "pcolormesh",
                "3d",
                "contour",
                "AllAtoms",
                "pdf",
                "cdf",
            ],
            help="additional parameter: 'withoutScatter' will NOT show scatter plot for 'xvg_box_compare'; 'imshow', 'pcolormesh', '3d', 'contour' were used for 'xpm_show' command; 'AllAtoms' were used for 'find_center' command; 'cdf' and 'pdf' are for 'xvg_show_distribution' command;",
        )
        parser.add_argument(
            "-al",
            "--additional_list",
            nargs="+",
            help="additional parameters. Used to set xtitles for 'xvg_ave_bar'",
        )
        parser.add_argument(
            "-ip",
            "--interpolation",
            type=str,
            default=None,
            help="specify the interpolation method, default to None",
        )
        parser.add_argument(
            "-ipf",
            "--interpolation_fold",
            type=int,
            default=10,
            help="specify the interpolation fold, default to 10",
        )

        args = parser.parse_args()
        self.__dict__ = args.__dict__
        self.__check_convert()

    def __parse_column(self, msg_line: str) -> List[int]:
        """parse column selections, generate List[int]"""
        lis: List[int] = []
        try:
            for msg in msg_line.strip(",").split(","):
                if msg.strip("-").count("-") == 2:
                    b, e = int(msg.split("-")[0]), int(msg.split("-")[1])
                    dt = int(msg.split("-")[2])
                    lis += [i for i in range(b, e, dt)]
                elif msg.strip("-").count("-") == 1:
                    b, e = int(msg.split("-")[0]), int(msg.split("-")[1])
                    lis += [i for i in range(b, e)]
                elif msg.strip("-").count("-") == 0:
                    lis.append(int(msg))
                else:
                    self.error(
                        f"wrong in parsing {msg}. '1-10', '1-10-2', or '10' were supported for input"
                    )
        except ValueError as verr:
            self.error(
                f"Error occured in parsing column selections, please input INTEGER ! \n {verr}"
            )
        except Exception as ex:
            self.error(f"Error occured in parsing column selections. {ex}")
        return lis

    def __check_convert(self) -> None:
        """check part of parameters and deal with parameter format"""
        ## deal parameters
        if self.input != None and "," in "".join(self.input):
            self.input = [fs.strip(",").split(",") for fs in self.input]
        column_select = []
        if self.columns != None:
            for columns in self.columns:
                column_select.append(self.__parse_column(columns))
        self.columns = column_select
        if self.legends != None and "," in "".join(self.legends):
            self.legends = [ls.strip(",").split(",") for ls in self.legends]
        ## check parameters
        if self.begin and self.begin < 0:
            self.warn(
                "parameter 'begin' should not be a minus, BE SURE for what you are doing"
            )
        if self.end and self.end < 0:
            self.warn(
                "parameter 'end' should not be a minus, BE SURE for what you are doing"
            )
        if self.dt and self.dt < 0:
            self.warn(
                "parameter 'dt' should not be a minus, BE SURE for what you are doing"
            )
        if self.x_precision and self.x_precision < 0:
            self.error("parameter 'x_precision' should not be a minus")
        if self.y_precision and self.y_precision < 0:
            self.error("parameter 'y_precision' should not be a minus")
        if self.z_precision and self.z_precision < 0:
            self.error("parameter 'z_precision' should not be a minus")



"""
xpmCommander module is part of DuIvyTools providing xpm related commands.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
from typing import List, Union

from scipy.interpolate import RectBivariateSpline, interp2d

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from Commands.Commands import Command
from FileParser.xpmParser import XPM
from utils import Parameters
from Visualizer.Visualizer_gnuplot import *
from Visualizer.Visualizer_matplotlib import *
from Visualizer.Visualizer_plotext import *
from Visualizer.Visualizer_plotly import *


class xpm_show(Command):
    """
    Visualize the xpm file.
    DIT support 4 plot engines (matplotlib, plotly, gnuplot, and plotext) and several modes to plot xpm into figures. 4 modes for matplotlib (imshow which is default, pcolormesh, 3d, and contour), and 3 modes for plotly and gnuplot (pcolormesh which is default, 3d, and contour). Plotext only support plotting simple and small size xpm in gray.
    Modes imshow and pcolormesh mainly show the matrix of xpm. For `Continuous` type xpms, matplotlib, plotly and gnuplot will NOT use its original colors, and the colormaps of each engine will be used. For matploblib and plotly, you can set colormaps by `-cmap`. For `Discrete` type of xpms, only pcolormesh of matploblib will NOT use its original colors. But you can set colors by mplstyle file or other style files.For the methods using its original colors, you can set colors by directly modifing the xpm file.
    Mode 3d mainly plot a 3d figure for `Continuous` xpm. Mode contour plot a contour figure for `Continuous` xpm. Also, you can set colormaps by `-cmap`.
    You can perform INTERPOLATION to data by specifing `-ip`.
    For imshow of matplotlib, the interpolation method was using the interpolation method of imshow function of matplobli, and there are lots of interpolation methods could be selected. If you do not know the names of interpolation methods, simply specify `-ip hhh`, then the error message will show you all names of interpolation methods for you to choose.
    For any other engines or modes, DIT use `scipy.interpolate.interp2d` to do the interpolation, so the methods for you to choose is `linear`, `cubic`, and `quintic`. Also, `-ip hhh` trick works. For this interpolation methods, you need to define a `--interpolation_fold` (default to 10).
    DIT support performing xpm cutting by `-xmin`, `-xmax`, `-ymin`, and `-ymax`, like only show 100*100 pixels from a 132*10000 DSSP xpm by setting `-xmin 100 -xmax 200 -ymin 200 -ymax 300`.

    :Parameters:
        -f, --input
                specify the input xpm file (or files)
        -o, --output (optional)
                specify the file name for saving figures
        -ns, --noshow (optional)
                whether not to show figures. When applied to gnuplot, DIT will generate a gnuplot input script
        -x, --xlabel (optional)
                specify the xlabel of figure
        -y, --ylabel (optional)
                specify the ylabel of figure
        -z, --zlabel (optional)
                specify the zlabel of figure
        -t, --title (optional)
                specify the title of figure
        -xs, --xshrink (optional)
                specify the shrink fold number of X values
        -ys, --yshrink (optional)
                specify the shrink fold number of Y values
        -zs, --zshrink (optional)
                specify the shrink fold number of Z values
        -xmin, --xmin (optional)
                specify the xmin index of xpm matrix to show
        -xmax, --xmax (optional)
                specify the xmax index of xpm matrix to show
        -ymin, --ymin (optional)
                specify the ymin index of xpm matrix to show
        -ymax, --ymax (optional)
                specify the ymax index of xpm matrix to show
        -zmin, --zmin (optional)
                specify the min value of colorbar to show
        -zmax, --zmax (optional)
                specify the max value of colorbar to show
        -m, --mode (optional)
                specify the mode of visualization: imshow, pcolormesh, 3d, contour
        -eg, --engine (optional)
                specify the plot engine: matplotlib (default), plotly, gnuplot, plotext
        -cmap, --colormap (optional)
                specify the colormap for visualization
        -ip, --interpolation (optional)
                specify the interpolation method
        -ipf, --interpolation_fold (optional)
                specify the multiple of interpolation
        --alpha (optional)
                specify the alpha of figure
        --x_precision (optional)
                specify the precision of X ticklabels
        --y_precision (optional)
                specify the precision of Y ticklabels
        --z_precision (optional)
                specify the precision of Z ticklabels
        --legend_location (optional)
                specify the location of legend, inside or outside
        --colorbar_location (optional)
                specify the location of colorbar, available for matplotlib: left, top, bottom, right

    :Usage:
        dit xpm_show -f FEL.xpm
        dit xpm_show -f hbond.xpm
        dit xpm_show -f DSSP.xpm -ns -o dssp.png
        dit xpm_show -f FEL.xpm -m pcolormesh -ip linear -ipf 5 -cmap solar
        dit xpm_show -f FEL.xpm -m 3d -x PC1 -y PC2 -z Energy -t FEL --alpha 0.5
        dit xpm_show -f FEL.xpm -m 3d --x_precision 1 --y_precision 2 --z_precision 0
        dit xpm_show -f FEL.xpm -m contour -cmap jet --colorbar_location bottom
        dit xpm_show -f FEL.xpm -m contour -cmap jet -zmin 0 -zmax 20
        dit xpm_show -f DSSP.xpm -xs 0.001 -x Time(ns) --legend_location outside
        dit xpm_show -f DSSP.xpm -eg plotly -xmin 1000 -xmax 2001 -ymin 50 -ymax 101
        dit xpm_show -f FEL.xpm -eg plotly -m 3d
        dit xpm_show -f FEL.xpm -eg plotly -m contour
        dit xpm_show -f DSSP.xpm -eg gnuplot --legend_location outside
        dit xpm_show -f FEL.xpm -eg gnuplot -m 3d -ip cubic
        dit xpm_show -f FEL.xpm -eg gnuplot -m contour -ns -o contour.png
    """

    def __init__(self, parm: Parameters) -> None:
        self.parm = parm

    def calc_interpolation(
        self,
        xaxis: List[float],
        yaxis: List[float],
        matrix: List[List[float]],
        method: str,
        ip_fold: int,
    ) -> Union[List[float], List[List[float]]]:
        """perform the interpolation of matrix data

        Args:
            xaxis (List[float]): the X data
            yaxis (List[float]): the Y data
            matrix (List[List[float]]): the matrix values
            method (str): method for interpolation
            ip_fold (int): the multiple of interpolation

        Returns:
            x_new (List[float]): the X data after interpolation
            y_new (List[float]): the Y data after interpolation
            matrix_new (List[List[float]]): the matrix values after interpolation
        """
        # interp2d : linear, cubic, quintic
        ip_func = interp2d(xaxis, yaxis, matrix, kind=method)
        # ip_func = RectBivariateSpline(xaxis, yaxis, matrix)#, kx=3, ky=3)
        x_new = np.linspace(np.min(xaxis), np.max(xaxis), ip_fold * len(xaxis))
        y_new = np.linspace(np.min(yaxis), np.max(yaxis), ip_fold * len(yaxis))
        matrix_new = ip_func(x_new, y_new)
        return x_new, y_new, matrix_new

    def hex2rgb(self, value):
        value = value.lstrip("#")
        res = tuple(int(value[i : i + 2], 16) for i in (0, 2, 4))
        return res

    def image_split(
        self, xaxis: List[float], yaxis: List[float], value_matrix: List[List[float]]
    ) -> Union[List[float], List[List[float]]]:
        """image cutting by xmin, xmax, ymin, and ymax

        Args:
            xaxis (List[float]): xaxis data
            yaxis (List[float]): yaxis data
            value_matrix (List[List[float]]): image value matrix

        Returns:
            xaxis (List[float]): xaxis data after cutting
            yaxis (List[float]): yaxis data after cutting
            value_matrix (List[List[float]]): image value matrix after cutting
        """

        if self.parm.xmin != None and not isinstance(self.parm.xmin, int):
            self.parm.xmin = int(self.parm.xmin)
        if self.parm.xmax != None and not isinstance(self.parm.xmax, int):
            self.parm.xmax = int(self.parm.xmax)
        if self.parm.ymin != None and not isinstance(self.parm.ymin, int):
            self.parm.ymin = int(self.parm.ymin)
        if self.parm.ymax != None and not isinstance(self.parm.ymax, int):
            self.parm.ymax = int(self.parm.ymax)

        if len(xaxis) != len(value_matrix[0]) or len(yaxis) != len(value_matrix):
            self.error(
                f"unequal size detected in image splitting: xaxis ({len(xaxis)}), yaxis ({len(yaxis)}), value_matrix ({len(value_matrix[0])}*{len(value_matrix)})"
            )

        if self.parm.xmin == None:
            self.parm.xmin = 0
        if self.parm.xmax == None:
            self.parm.xmax = len(xaxis)
        if self.parm.ymin == None:
            self.parm.ymin = 0
        if self.parm.ymax == None:
            self.parm.ymax = len(yaxis)

        xmin, xmax = self.parm.xmin, self.parm.xmax
        ymin, ymax = self.parm.ymin, self.parm.ymax
        xaxis = xaxis[xmin:xmax]
        for y, _ in enumerate(yaxis):
            value_matrix[y] = value_matrix[y][xmin:xmax]
        value_matrix.reverse()  # from bottom to top
        yaxis.reverse()
        value_matrix = value_matrix[ymin:ymax]
        yaxis = yaxis[ymin:ymax]
        value_matrix.reverse()  # from top to bottom
        yaxis.reverse()

        if len(xaxis) < 1 or len(yaxis) < 1:
            self.error(
                f"the image ({len(xaxis)}*{len(yaxis)}) less than 1 dimensions, unable to draw"
            )
        self.info(
            f"cutting image by x_index in range [{xmin}, {xmax}), y_index in range [{ymin}, {ymax})"
        )
        self.info(
            f"cutting image by xaxis in range [{xaxis[0]}, {xaxis[-1]}], yaxis in range [{yaxis[-1]}, {yaxis[0]}]"
        )

        return xaxis, yaxis, value_matrix

    def __call__(self):
        # self.info("in xpm_show")
        # print(self.parm.__dict__)

        if not self.parm.input:
            self.error("you must specify a xpm file to show")

        for xpmfile in self.parm.input:
            xpm = XPM(xpmfile)
            self.file = xpm
            self.remove_latex(filetype="XPM")

            xaxis = [x * self.parm.xshrink for x in xpm.xaxis]
            yaxis = [y * self.parm.yshrink for y in xpm.yaxis]
            value_matrix = []
            for y, _ in enumerate(yaxis):
                v_lis = []
                for x, _ in enumerate(xaxis):
                    v_lis.append(xpm.value_matrix[y][x] * self.parm.zshrink)
                value_matrix.append(v_lis)

            ## top -> bottom ===>>> bottom to top
            yaxis.reverse()
            value_matrix.reverse()

            xaxis, yaxis, value_matrix = self.image_split(xaxis, yaxis, value_matrix)

            kwargs = {
                "data_list": value_matrix,
                "xdata_list": xaxis,
                "ydata_list": yaxis,
                "legends": self.sel_parm(self.parm.legends, xpm.notes),
                "color_list": xpm.colors,
                "zmin": self.parm.zmin,
                "zmax": self.parm.zmax,
                "xlabel": self.get_parm("xlabel"),
                "ylabel": self.get_parm("ylabel"),
                "zlabel": self.sel_parm(self.parm.zlabel, xpm.legend),
                "title": self.get_parm("title"),
                "x_precision": self.parm.x_precision,
                "y_precision": self.parm.y_precision,
                "z_precision": self.parm.z_precision,
                "alpha": self.parm.alpha,
                "legend_location": self.sel_parm(self.parm.legend_location, "outside"),
                "colorbar_location": self.parm.colorbar_location,
                "fig_type": xpm.type,
                "cmap": self.parm.colormap,
            }

            interpolation = self.parm.interpolation
            ip_fold = self.parm.interpolation_fold
            mode = self.parm.mode

            if self.parm.engine == "matplotlib":
                if mode in ["pcolormesh", "3d", "contour"]:
                    if interpolation != None:
                        if self.file.type != "Continuous":
                            self.warn(
                                f"you are applying interpolation to {self.file.type} type of XPM. It should not be, but DIT would do it. BE CAREFUL for what you get !"
                            )
                        xaxis, yaxis, value_matrix = self.calc_interpolation(
                            xaxis, yaxis, value_matrix, interpolation, ip_fold
                        )
                        kwargs["xdata_list"] = xaxis
                        kwargs["ydata_list"] = yaxis
                        kwargs["data_list"] = value_matrix
                    else:
                        kwargs["xdata_list"] = np.array(xaxis)
                        kwargs["ydata_list"] = np.array(yaxis)
                        kwargs["data_list"] = np.array(kwargs["data_list"])
                if mode == "pcolormesh":
                    fig = PcolormeshMatplotlib(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)
                elif mode == "3d":
                    fig = ThreeDimensionMatplotlib(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)
                elif mode == "contour":
                    fig = ContourMatplotlib(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)
                else:
                    kwargs["interpolation"] = interpolation
                    fig = ImshowMatplotlib(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)

            elif self.parm.engine == "plotly":
                if interpolation != None:
                    if self.file.type != "Continuous":
                        self.warn(
                            f"you are applying interpolation to {self.file.type} type of XPM. It should not be, but DIT would do it. BE CAREFUL for what you get !"
                        )
                    xaxis, yaxis, value_matrix = self.calc_interpolation(
                        xaxis, yaxis, value_matrix, interpolation, ip_fold
                    )
                    kwargs["xdata_list"] = xaxis
                    kwargs["ydata_list"] = yaxis
                    kwargs["data_list"] = value_matrix
                if mode == "3d":
                    fig = ThreeDimensionPlotly(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)
                elif mode == "contour":
                    fig = ContourPlotly(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)
                else:
                    fig = PcolormeshPlotly(**kwargs)
                    fig.final(self.parm.output, self.parm.noshow)

            elif self.parm.engine == "gnuplot":
                if interpolation != None:
                    if self.file.type != "Continuous":
                        self.warn(
                            f"you are applying interpolation to {self.file.type} type of XPM. It should not be, but DIT would do it. BE CAREFUL for what you get !"
                        )
                    xaxis, yaxis, value_matrix = self.calc_interpolation(
                        xaxis, yaxis, value_matrix, interpolation, ip_fold
                    )
                    kwargs["xdata_list"] = xaxis
                    kwargs["ydata_list"] = yaxis
                    kwargs["data_list"] = value_matrix
                if mode not in ["3d", "contour"]:
                    mode = "imshow"
                fig = ImshowGnuplot(mode, **kwargs)
                fig.final(self.parm.output, self.parm.noshow)

            elif self.parm.engine == "plotext":
                if interpolation != None:
                    self.warn(
                        "plotext engine do not support interpolation now, ignored it"
                    )
                color_matrix = []
                for y, _ in enumerate(yaxis):
                    c_lis = []
                    for x, _ in enumerate(xaxis):
                        c = xpm.colors[xpm.chars.index(xpm.dot_matrix[y][x])]
                        c_lis.append(self.hex2rgb(c))
                    color_matrix.append(c_lis)
                kwargs["data_list"] = color_matrix
                ImshowPlotext(**kwargs)
            else:
                self.error("wrong selection of plot engine")


class xpm2csv(Command):
    """
    Converting xpm file into csv file in (X, Y, Z) data format.

    :Parameters:
        -f, --input
                specify the input xpm file (or files)
        -o, --output (optional)
                specify the csv file name for saving data, default to 'dit_{input}.csv'
        -x, --xlabel (optional)
                specify the xlabel of XPM
        -y, --ylabel (optional)
                specify the ylabel of XPM
        -z, --zlabel (optional)
                specify the zlabel of XPM
        -xs, --xshrink (optional)
                specify the shrink fold number of X values
        -ys, --yshrink (optional)
                specify the shrink fold number of Y values
        -zs, --zshrink (optional)
                specify the shrink fold number of Z values

    :Usage:
        dit xpm2csv -f FEL.xpm -o fel.csv
        dit xpm2csv -f DSSP.xpm -o dssp.csv -x Time(ns) -xs 0.001 -y "Residue No."
    """

    def __init__(self, parm: Parameters) -> None:
        self.parm = parm

    def __call__(self):
        # self.info("in xpm2csv")
        # print(self.parm.__dict__)

        if not self.parm.input:
            self.error("you must specify a xpm file for converting to csv file")
        if len(self.parm.input) > 1:
            self.warn(f"only the first xpm file {self.parm.input[0]} will be used !")
        if not self.parm.output:
            self.parm.output = f"dit_{self.parm.input[0].split('.')[0]}.csv"
        self.parm.output = self.check_output_exist(self.parm.output)

        xpm = XPM(self.parm.input[0])
        with open(self.parm.output, "w") as fo:
            if xpm.type != "Continuous":
                fo.write("#### DIT: it's not a Continuous type xpm file\n")
                fo.write("#### labels of values were recorded below: \n")
                for index, label in enumerate(xpm.notes):
                    fo.write(f"#### {index} : {label}\n")
            x_title = (xpm.xlabel, "x-axis")[len(xpm.xlabel) == 0]
            y_title = (xpm.ylabel, "y-axis")[len(xpm.ylabel) == 0]
            z_title = (xpm.legend, "value")[len(xpm.legend) == 0]
            x_title = self.sel_parm(self.parm.xlabel, x_title)
            y_title = self.sel_parm(self.parm.ylabel, y_title)
            z_title = self.sel_parm(self.parm.zlabel, z_title)
            fo.write(f"{x_title},{y_title},{z_title}\n")
            for y, y_value in enumerate(xpm.yaxis):
                for x, x_value in enumerate(xpm.xaxis):
                    z_value = xpm.value_matrix[y][x]
                    x_value *= self.parm.xshrink
                    y_value *= self.parm.yshrink
                    z_value *= self.parm.zshrink
                    fo.write(f"{x_value:.6f},{y_value:.6f},{z_value:.6f}\n")
        self.info(
            f"extract data from {xpm.xpmfile} and saved into {self.parm.output} successfully"
        )


class xpm2dat(Command):
    """
    Converting xpm file into dat file in data matrix format.

    :Parameters:
        -f, --input
                specify the input xpm file (or files)
        -o, --output (optional)
                specify the dat file name for saving data, default to 'dit_{input}.dat'
        -x, --xlabel (optional)
                specify the xlabel of XPM
        -y, --ylabel (optional)
                specify the ylabel of XPM
        -z, --zlabel (optional)
                specify the zlabel of XPM
        -xs, --xshrink (optional)
                specify the shrink fold number of X values
        -ys, --yshrink (optional)
                specify the shrink fold number of Y values
        -zs, --zshrink (optional)
                specify the shrink fold number of Z values

    :Usage:
        dit xpm2dat -f FEL.xpm -o fel.dat
        dit xpm2dat -f DSSP.xpm -o dssp.dat -x Time(ns) -xs 0.001
    """

    def __init__(self, parm: Parameters) -> None:
        self.parm = parm

    def __call__(self):
        # self.info("in xpm2dat")
        # print(self.parm.__dict__)

        if not self.parm.input:
            self.error("you must specify a xpm file for converting to dat file")
        if len(self.parm.input) > 1:
            self.warn("only the first xpm file you specified will be used !")
        if not self.parm.output:
            self.parm.output = f"dit_{self.parm.input[0].split('.')[0]}.dat"
        self.parm.output = self.check_output_exist(self.parm.output)

        xpm = XPM(self.parm.input[0])
        with open(self.parm.output, "w") as fo:
            if xpm.type != "Continuous":
                fo.write("#### DIT: it's not a Continuous type xpm file\n")
                fo.write("#### labels of values were recorded below: \n")
                for index, label in enumerate(xpm.notes):
                    fo.write(f"#### {index} : {label}\n")
            x_title = (xpm.xlabel, "x-axis")[len(xpm.xlabel) == 0]
            y_title = (xpm.ylabel, "y-axis")[len(xpm.ylabel) == 0]
            z_title = (xpm.legend, "value")[len(xpm.legend) == 0]
            x_title = self.sel_parm(self.parm.xlabel, x_title)
            y_title = self.sel_parm(self.parm.ylabel, y_title)
            z_title = self.sel_parm(self.parm.zlabel, z_title)
            fo.write(
                "#### "
                + f"{x_title} (xaxis) data were shown below, from left to right:\n"
            )
            fo.write(",".join([f"{x*self.parm.xshrink:.6f}" for x in xpm.xaxis]) + "\n")
            fo.write(
                "#### "
                + f"{y_title} (yaxis) data were shown below, from top to bottom:\n"
            )
            fo.write(",".join([f"{y*self.parm.yshrink:.6f}" for y in xpm.yaxis]) + "\n")
            fo.write(
                "#### "
                + f"{y_title} (yaxis) data were shown below, from bottom to top:\n"
            )
            y_lis = [f"{y*self.parm.yshrink:.6f}" for y in xpm.yaxis]
            y_lis.reverse()
            fo.write(",".join(y_lis) + "\n")
            fo.write(
                "#### "
                + f"{z_title} (figure dots) data were shown below, from top to bottom, from left to right:\n"
            )
            for y, _ in enumerate(xpm.yaxis):
                line_list: List[str] = []
                for x, _ in enumerate(xpm.xaxis):
                    z_value = xpm.value_matrix[y][x]
                    line_list.append(f"{z_value*self.parm.zshrink:.6f}")
                fo.write(",".join(line_list) + "\n")
        self.info(
            f"extract data from {xpm.xpmfile} and saved into {self.parm.output} successfully"
        )


class xpm_diff(Command):
    """
    Calculate the difference of two xpm files.
    By calculating the difference of matrix values of two xpms correspondingly (first xpm - second xpm), this command could be used for presentation the variation of two xpm files.

    :Parameters:
        -f, --input
                specify the two xpm files for input
        -o, --output (optional)
                specify the xpm file name for output, default to 'dit_xpm_diff.xpm'
        -x, --xlabel (optional)
                specify the xlabel of XPM of output
        -y, --ylabel (optional)
                specify the ylabel of XPM of output
        -z, --zlabel (optional)
                specify the zlabel of XPM of output
        -xs, --xshrink (optional)
                specify the shrink fold number of X values
        -ys, --yshrink (optional)
                specify the shrink fold number of Y values
        -zs, --zshrink (optional)
                specify the shrink fold number of Z values

    :Usage:
        dit xpm_diff -f DCCM0.xpm DCCM1.xpm -o DCCM0-1.xpm
    """

    def __init__(self, parm: Parameters) -> None:
        self.parm = parm

    def __call__(self):
        # self.info("in xpm_diff")
        # print(self.parm.__dict__)

        if not self.parm.input:
            self.error("you must specify two xpm file for calcuelation")
        if len(self.parm.input) > 2:
            self.warn("only the first two xpm file you specified will be used !")
        if len(self.parm.input) < 2:
            self.warn("at least two xpm file are needed for calculation !")
        if not self.parm.output:
            self.parm.output = "dit_xpm_diff.xpm"
        self.parm.output = self.check_output_exist(self.parm.output)

        xpm0 = XPM(self.parm.input[0])
        xpm1 = XPM(self.parm.input[1])
        xpm = xpm0 - xpm1
        xpm.xpmfile = self.parm.output
        xpm.xlabel = self.sel_parm(self.parm.xlabel, xpm.xlabel)
        xpm.ylabel = self.sel_parm(self.parm.ylabel, xpm.ylabel)
        xpm.legend = self.sel_parm(self.parm.zlabel, xpm.legend)
        xpm.title = self.sel_parm(self.parm.title, f"{xpm0.xpmfile} - {xpm1.xpmfile}")
        xpm.xaxis = [x * self.parm.xshrink for x in xpm.xaxis]
        xpm.yaxis = [y * self.parm.yshrink for y in xpm.yaxis]
        for y, _ in enumerate(xpm.yaxis):
            for x, _ in enumerate(xpm.xaxis):
                xpm.value_matrix[y][x] *= self.parm.zshrink
        xpm.save(self.parm.output)


class xpm_merge(Command):
    """
    Merge two xpm files half by half.
    The first specified xpm will be located at left top corner, the second one will be located at right bottom corner. Thi command might be useful for stitching two xpm together.

    :Parameters:
        -f, --input
                specify the two xpm files for input
        -o, --output (optional)
                specify the xpm file name for output, default to 'dit_xpm_merge.xpm'
        -x, --xlabel (optional)
                specify the xlabel of XPM of output
        -y, --ylabel (optional)
                specify the ylabel of XPM of output
        -z, --zlabel (optional)
                specify the zlabel of XPM of output
        -xs, --xshrink (optional)
                specify the shrink fold number of X values
        -ys, --yshrink (optional)
                specify the shrink fold number of Y values
        -zs, --zshrink (optional)
                specify the shrink fold number of Z values

    :Usage:
        dit xpm_merge -f DCCM0.xpm DCCM1.xpm -o DCCM0-1.xpm
    """

    def __init__(self, parm: Parameters) -> None:
        self.parm = parm

    def __call__(self):
        # self.info("in xpm_merge")
        # print(self.parm.__dict__)

        if not self.parm.input:
            self.error("you must specify two xpm file for calcuelation")
        if len(self.parm.input) > 2:
            self.warn("only the first two xpm file you specified will be used !")
        if len(self.parm.input) < 2:
            self.warn("at least two xpm file are needed for calculation !")
        if not self.parm.output:
            self.parm.output = "dit_xpm_merge.xpm"
        self.parm.output = self.check_output_exist(self.parm.output)

        xpm0 = XPM(self.parm.input[0])
        xpm1 = XPM(self.parm.input[1])
        for key in ["title", "xlabel", "ylabel", "xaxis", "yaxis"]:
            if xpm0.__dict__[key] != xpm1.__dict__[key]:
                self.warn(
                    f"Detected different {key} in {xpm0.xpmfile} and {xpm1.xpmfile}. \nDIT strongly warns you that different type (meanings) of xpms should NOT be used to merge. The results may NOT be reliable !!! "
                )
        if xpm0.type != xpm1.type:
            self.error(
                f"Do not support different types of xpm to merge:{xpm0.xpmfile}({xpm0.type}), and {xpm1.xpmfile}({xpm1.type})"
            )
        if xpm0.width != xpm1.width or xpm0.height != xpm1.height:
            self.error(
                f"The shape of {xpm0.xpmfile} ({xpm0.width}, {xpm0.height}) and {xpm1.xpmfile} ({xpm1.width}, {xpm1.height}) are different, unable to calculate difference."
            )

        out = XPM(self.parm.output, is_file=False, new_file=True)
        for key, value in xpm0.__dict__.items():
            out.__dict__[key] = value
        if out.type == "Continuous":
            for h in range(xpm0.height):
                for w in range(xpm0.width):
                    if h / xpm0.height + w / xpm0.width < 1:
                        value = xpm0.value_matrix[h][w]  # left top
                    else:
                        value = xpm1.value_matrix[h][w]  # right bottom
                    out.value_matrix[h][w] = value
            out.refresh_by_value_matrix()
        else:
            for h in range(xpm0.height):
                for w in range(xpm0.width):
                    if h / xpm0.height + w / xpm0.width < 1:
                        value = xpm0.notes[xpm0.value_matrix[h][w]]  # left top
                    else:
                        value = xpm1.notes[xpm1.value_matrix[h][w]]  # right bottom
                    out.value_matrix[h][w] = value
            out.refresh_by_value_matrix(is_Continuous=False)

        out.title = self.sel_parm(
            self.parm.title, f"{xpm0.xpmfile}(left) / {xpm1.xpmfile}(right)"
        )
        out.xlabel = self.sel_parm(self.parm.xlabel, out.xlabel)
        out.ylabel = self.sel_parm(self.parm.ylabel, out.ylabel)
        out.legend = self.sel_parm(self.parm.zlabel, out.legend)
        out.xaxis = [x * self.parm.xshrink for x in out.xaxis]
        out.yaxis = [y * self.parm.yshrink for y in out.yaxis]
        for y, _ in enumerate(out.yaxis):
            for x, _ in enumerate(out.xaxis):
                out.value_matrix[y][x] *= self.parm.zshrink
        out.save(self.parm.output)



"""
xpmParser module is part of DuIvyTools for parsing the xpm file generated by GROMACS.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import string
from itertools import chain
from typing import List

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log


class XPM(log):
    """XPM class was designed to parse xpm file"""

    def __init__(
        self, xpmfile: str, is_file: bool = True, new_file: bool = False
    ) -> None:
        self.title: str = ""
        self.legend: str = ""
        self.type: str = ""
        self.xlabel: str = ""
        self.ylabel: str = ""
        self.width: int = 0
        self.height: int = 0
        self.color_num: int = 0
        self.char_per_pixel: int = 0
        self.chars: list[str] = []
        self.colors: list[str] = []
        self.notes: list[str] = []
        self.xaxis: list[float] = []
        self.yaxis: list[float] = []
        self.datalines: list[str] = []
        self.dot_matrix: list[list[str]] = []
        self.value_matrix: list[list[float]] = []

        if new_file:
            self.xpmfile = xpmfile
        else:
            if is_file:
                self.xpmfile: str = xpmfile
                if not os.path.exists(xpmfile):
                    self.error(f"No {xpmfile} detected ! check it !")
                if xpmfile[-4:] != ".xpm":
                    self.error(
                        f"you must specify a file with suffix .xpm, instead of {xpmfile}"
                    )
                with open(xpmfile, "r") as fo:
                    content = fo.read()
            else:
                content = xpmfile
            lines = [l.strip() for l in content.strip().split("\n")]
            self.parse_xpm(lines)
            if is_file:
                self.info(f"parsing data from {xpmfile} successfully !")

    def parse_xpm(self, lines: List[str]) -> None:
        """convert xpm content (lines) into XPM class

        Args:
            lines (List[str]): xpm file lines
        """
        flag_4_code: int = 0
        for line in lines:
            if flag_4_code == 1:
                flag_4_code = 2
                code4 = [int(c) for c in line.strip(' ,"').split()]
                self.width = code4[0]
                self.height = code4[1]
                self.color_num = code4[2]
                self.char_per_pixel = code4[3]
                continue
            elif (flag_4_code == 0) and line.startswith("static char"):
                flag_4_code = 1
                continue

            ## parsing comments and axis parts
            if line.startswith("/* x-axis"):
                self.xaxis += [float(n) for n in line.strip().split()[2:-1]]
                continue
            elif line.startswith("/* y-axis"):
                self.yaxis += [float(n) for n in line.strip().split()[2:-1]]
                continue
            elif line.startswith("/* title"):
                self.title = line.strip().split('"')[1]
                continue
            elif line.startswith("/* legend"):
                self.legend = line.strip().split('"')[1]
                continue
            elif line.startswith("/* x-label"):
                self.xlabel = line.strip().split('"')[1]
                continue
            elif line.startswith("/* y-label"):
                self.ylabel = line.strip().split('"')[1]
                continue
            elif line.startswith("/* type"):
                self.type = line.strip().split('"')[1]
                continue

            items = line.strip().split()
            ## for char-color-note part
            if len(items) == 7 and items[1] == "c":
                self.colors.append(items[2])
                self.notes.append(items[5].strip('"'))
                if len(items[0].strip('"')) == self.char_per_pixel:
                    self.chars.append(items[0].strip('"'))
                ## deal with blank char
                if len(items[0].strip('"')) < self.char_per_pixel:
                    char_item = line.strip('"')[: self.char_per_pixel]
                    self.chars.append(char_item)
                continue

            ## for figure content part
            if line.strip().startswith('"') and (
                len(line.strip().strip(",").strip('"'))
                == self.width * self.char_per_pixel
            ):
                self.datalines.append(line.strip().strip(",").strip('"'))

        ## check infos
        if self.color_num != len(self.chars):
            self.error(f"length of chars != color_num")
        if self.color_num != len(self.colors):
            self.error(f"length of colors != color_num")
        if self.color_num != len(self.notes):
            self.error(f"length of notes != color_num")
        if len(self.xaxis) != self.width and len(self.xaxis) != self.width + 1:
            self.error(
                f"length of xaxis ({len(self.xaxis)}) != xpm width ({self.width}) or xpm width +1"
            )
        if len(self.yaxis) != self.height and len(self.yaxis) != self.height + 1:
            self.error(
                f"length of yaxis ({len(self.yaxis)}) != xpm height ({self.height}) or xpm height +1"
            )
        if len(self.datalines) != self.height:
            self.error(
                f"rows of data ({len(self.datalines)}) is not equal to xpm height ({self.height}), check it!"
            )

        ## convert data
        if len(self.xaxis) == self.width + 1:
            self.xaxis = [
                (self.xaxis[i - 1] + self.xaxis[i]) / 2.0
                for i in range(1, len(self.xaxis))
            ]
            self.warn(
                "length of x-axis is 1 more than xpm width, use intermediate value for instead. "
            )
        if len(self.yaxis) == self.height + 1:
            self.yaxis = [
                (self.yaxis[i - 1] + self.yaxis[i]) / 2.0
                for i in range(1, len(self.yaxis))
            ]
            self.warn(
                "length of y-axis is 1 more than xpm height, use intermediate value for instead. "
            )
        if self.type == "Continuous":
            self.notes = [float(n) for n in self.notes]

        self.yaxis.reverse()  # IMPORTANT! from high to low now

        for dataline in self.datalines:
            dot_list: List[str] = []
            value_list: List[float] = []
            for i in range(0, self.width * self.char_per_pixel, self.char_per_pixel):
                dot = dataline[i : i + self.char_per_pixel]
                dot_list.append(dot)
                if self.type == "Continuous":
                    value = self.notes[self.chars.index(dot)]
                else:
                    # for Discrete, value store the index of chars|notes|colors
                    value = self.chars.index(dot)
                value_list.append(value)
            self.dot_matrix.append(dot_list)
            self.value_matrix.append(value_list)

        if (
            len(self.dot_matrix) != self.height
            or len(self.value_matrix) != self.height
            or len(self.dot_matrix[0]) != self.width
            or len(self.value_matrix[0]) != self.width
        ):
            self.error("Dimension error while parsing xpm file")

    def __sub__(self, xpm):  # diff_map
        """values of self - values of xpm correspondingly, return a new result XPM"""
        if self.type != "Continuous" or xpm.type != "Continuous":
            self.error("Only supported to apply xpm_diff to Continuous type of xpms")
        for key in ["title", "xlabel", "ylabel", "xaxis", "yaxis"]:
            if self.__dict__[key] != xpm.__dict__[key]:
                self.warn(
                    f"Detected different {key} in {self.xpmfile} and {xpm.xpmfile}. \nDIT strongly warns you that different type (meanings) of xpms should NOT be used to calculate difference. The results would NOT be reliable !!! "
                )
        if self.width != xpm.width or self.height != xpm.height:
            self.error(
                f"The shape of {self.xpmfile} ({self.width}, {self.height}) and {xpm.xpmfile} ({xpm.width}, {xpm.height}) are different, unable to calculate difference."
            )

        out = XPM("", is_file=False, new_file=True)
        for key, value in self.__dict__.items():
            out.__dict__[key] = value
        for h in range(self.height):
            for w in range(self.width):
                value = self.value_matrix[h][w] - xpm.value_matrix[h][w]
                value = float(f"{value:.6f}")
                out.value_matrix[h][w] = value
        out.refresh_by_value_matrix()
        return out

    def refresh_by_value_matrix(self, is_Continuous: bool = True) -> None:
        """generate the xpm class content by value_matrix

        Args:
            is_Continuous (bool, optional): set the type of xpm. Defaults to True.
        """
        out_value_list = sorted(list(set(chain(*self.value_matrix))))
        self.notes = out_value_list
        colors, l = [], len(out_value_list)
        for i in range(l):
            c = 16000000 // len(out_value_list) * i
            colors.append(f'#{hex(c)[-6:].replace("x", "0").upper():0>6}')
        self.color_num = l
        self.colors = colors

        letters = string.ascii_letters + "0123456789!@#$%^&*()-_=+{}|;"
        if l <= 0:
            self.error("no data detected in xpm.value_matrix")
        elif l > 0 and l <= len(letters):
            self.char_per_pixel = 1
            self.chars = [f"{x}" for x in letters][:l]
        elif l > len(letters) and l <= len(letters) * len(letters):
            self.char_per_pixel = 2
            self.chars = [f"{x}{y}" for x in letters for y in letters][:l]
        elif l > len(letters) * len(letters) and l <= len(letters) * len(letters) * len(
            letters
        ):
            self.warn(
                f"so many values ({l}) in xpm.value_matrix, may resulting in large xpm file"
            )
            self.char_per_pixel = 3
            self.chars = [
                f"{x}{y}{z}" for x in letters for y in letters for z in letters
            ][:l]
        else:
            self.error(
                f"too many values ({l}) in xpm.value_matrix, only able to construct XPM with {len(letters)*len(letters)*len(letters)} values or less"
            )

        if len(self.dot_matrix) == 0 or len(self.datalines) == 0:
            for h in range(self.height):
                self.dot_matrix.append(["" for _ in range(self.width)])
                self.datalines.append("")
        for h in range(self.height):
            dot_line: str = ""
            for w in range(self.width):
                dot = self.chars[out_value_list.index(self.value_matrix[h][w])]
                self.dot_matrix[h][w] = dot
                dot_line += dot
            self.datalines[h] = dot_line
        if not is_Continuous:  # refresh value_matrix from str to index
            for h in range(self.height):
                for w in range(self.width):
                    self.value_matrix[h][w] = out_value_list.index(
                        self.value_matrix[h][w]
                    )

    def save(self, outname: str) -> None:
        """dump XPM into xpm file"""
        outstr = "/* XPM */\n/* This file is generated by DuIvyTools */\n"
        outstr += f"""/* title: "{self.title}" */\n"""
        outstr += f"""/* legend: "{self.legend}" */\n"""
        outstr += f"""/* x-label: "{self.xlabel}" */\n"""
        outstr += f"""/* y-label: "{self.ylabel}" */\n"""
        outstr += f"""/* type: "{self.type}" */\n"""
        outstr += """static char *gromacs_xpm[] = {\n"""
        outstr += f""""{self.width} {self.height} {self.color_num} {self.char_per_pixel}",\n"""
        for char, color, note in zip(self.chars, self.colors, self.notes):
            outstr += f""""{char}  c {color} " /* "{note}" */,\n"""
        for x in range(0, len(self.xaxis), 50):
            outstr += (
                "/* x-axis: "
                + " ".join(str(v) for v in self.xaxis[x : x + 50])
                + " */\n"
            )
        # should not change original xpm data
        yaxis = [y for y in reversed(self.yaxis)]  # top to bottom -> bottom to top
        for y in range(0, len(yaxis), 50):
            outstr += (
                "/* y-axis: " + " ".join(str(v) for v in yaxis[y : y + 50]) + " */\n"
            )
        for line in self.datalines:
            outstr += f""""{line}",\n"""
        outstr = outstr.strip().strip(",") + "\n"
        with open(outname, "w") as fo:
            fo.write(outstr)
        self.info(f"Save results into {outname} successfully")


class XPMS(log):
    """XPMS class was designed to parse xpm file with multi-frames"""

    def __init__(self, xpmfile: str) -> None:
        self.xpmfile: str = xpmfile
        self.frames: list[XPM] = []

        with open(xpmfile, "r") as fo:
            contents = fo.read()
        contents = contents.split("/* XPM */")
        contents = [f"/* XPM */\n{c}" for c in contents if c.strip() != ""]
        for content in contents:
            xpm = XPM(content, is_file=False)
            self.frames.append(xpm)
        self.info(f"parsing multi-frames data from {xpmfile} successfully !")

    def __len__(self) -> int:
        """return number of frames"""
        return len(self.frames)

    def __getitem__(self, index: int) -> XPM:
        """get XPM of one frame"""
        return self.frames[index]

    def get_time_series(self) -> List[float]:
        """parsing time infos from xpm titles"""
        times: float = []
        for xpm in self:
            if xpm.title.startswith("t="):
                times.append(int(xpm.title[2:-2]))
            else:
                self.error(f"cannot parse time info from xpm title {xpm.title}")
        return times



"""
This module provides four plot engines (matplotlib, plotly, gnuplot, and plotext) and several type of plots, including scatter, line, heatmap, and so on. 

This package is provided to you under GPLv3 License.
"""


"""
Pipi_dist_ang module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to calculate the distance ang angles between two rings you 
specified in index file or between a ring and a vector

This module contains:
    - pipi_dist_ang function

This file is provided to you by GPLv3 license."""


import os
import sys
import math
import argparse
import logging


def calcDist(ring_1_frames: list = [], ring_2_frames: list = []) -> list:
    """
    calcDist: to calculate the distances of two rings

    :parameter:
        ring_1_frames: a list stored coordinates of ring 1 of different frames
        ring_2_frames: a list stored coordinates of ring 2 of different frames

    :return:
        distance: a list to store distance
    """

    ## check the frames of two ring
    if len(ring_1_frames) != len(ring_2_frames):
        logging.error("length of frames of coordinates isn't equal")
        sys.exit()
    ## calculate the distance between two ring
    distance = []
    for i in range(len(ring_1_frames)):
        ring_1_coor = ring_1_frames[i]
        ring_2_coor = ring_2_frames[i]
        ring_1_atom_num = len(ring_1_coor) * 1.0
        ring_2_atom_num = len(ring_2_coor) * 1.0
        ## calculate the center of two ring
        ring_1_center = [
            sum([coor[0] for coor in ring_1_coor]) / ring_1_atom_num,
            sum([coor[1] for coor in ring_1_coor]) / ring_1_atom_num,
            sum([coor[2] for coor in ring_1_coor]) / ring_1_atom_num,
        ]
        ring_2_center = [
            sum([coor[0] for coor in ring_2_coor]) / ring_2_atom_num,
            sum([coor[1] for coor in ring_2_coor]) / ring_2_atom_num,
            sum([coor[2] for coor in ring_2_coor]) / ring_2_atom_num,
        ]
        ## calculate the distance
        dist = (
            (ring_1_center[0] - ring_2_center[0]) ** 2
            + (ring_1_center[1] - ring_2_center[1]) ** 2
            + (ring_1_center[2] - ring_2_center[2]) ** 2
        ) ** 0.5
        distance.append(dist)

    return distance


def calcAng(ring_1_frames: list = [], ring_2_frames: list = []) -> list:
    """
    calcAng: to calculate the Angles of two rings

    :parameter:
        ring_1_frames: a list stored coordinates of ring 1 of different frames
        ring_2_frames: a list stored coordinates of ring 2 of different frames

    :return:
        angle: a list to store angles
    """
    ## check the frames of two ring
    if len(ring_1_frames) != len(ring_2_frames):
        logging.error("length of frames of coordinates isn't equal")
        exit()
    ## calculate the angles
    angles = []
    for i in range(len(ring_1_frames)):
        ring_1_coor = ring_1_frames[i]
        ring_2_coor = ring_2_frames[i]
        ## p1 = [ a1, a2, a3 ], p2 = [ b1, b2, b3 ]
        ## p1xp2 = [ a2b3 - a3b2, a3b1 - a1b3, a1b2 - a2b1 ]
        ## calculate the plane normal of ring 1 by cross product
        r1_p1 = [
            ring_1_coor[2][0] - ring_1_coor[0][0],
            ring_1_coor[2][1] - ring_1_coor[0][1],
            ring_1_coor[2][2] - ring_1_coor[0][2],
        ]
        r1_p2 = [
            ring_1_coor[4][0] - ring_1_coor[0][0],
            ring_1_coor[4][1] - ring_1_coor[0][1],
            ring_1_coor[4][2] - ring_1_coor[0][2],
        ]
        r1_xProd = [
            r1_p1[1] * r1_p2[2] - r1_p1[2] * r1_p2[1],
            r1_p1[2] * r1_p2[0] - r1_p1[0] * r1_p2[2],
            r1_p1[0] * r1_p2[1] - r1_p1[1] * r1_p2[0],
        ]
        ## calculate the plane normal of ring 2 by cross product
        r2_p1 = [
            ring_2_coor[2][0] - ring_2_coor[0][0],
            ring_2_coor[2][1] - ring_2_coor[0][1],
            ring_2_coor[2][2] - ring_2_coor[0][2],
        ]
        r2_p2 = [
            ring_2_coor[4][0] - ring_2_coor[0][0],
            ring_2_coor[4][1] - ring_2_coor[0][1],
            ring_2_coor[4][2] - ring_2_coor[0][2],
        ]
        r2_xProd = [
            r2_p1[1] * r2_p2[2] - r2_p1[2] * r2_p2[1],
            r2_p1[2] * r2_p2[0] - r2_p1[0] * r2_p2[2],
            r2_p1[0] * r2_p2[1] - r2_p1[1] * r2_p2[0],
        ]
        ## calc the degree
        dotProduct = r1_xProd[0] * r2_xProd[0] + r1_xProd[1] * r2_xProd[1]
        dotProduct += r1_xProd[2] * r2_xProd[2]
        r1_xProd_norm = (r1_xProd[0] ** 2 + r1_xProd[1] ** 2 + r1_xProd[2] ** 2) ** 0.5
        r2_xProd_norm = (r2_xProd[0] ** 2 + r2_xProd[1] ** 2 + r2_xProd[2] ** 2) ** 0.5
        cos_degree = dotProduct / (r1_xProd_norm * r2_xProd_norm)
        degree = math.acos(cos_degree) * 180 / math.pi
        if degree > 90:
            degree = 180 - degree
        angles.append(degree)

    return angles


def getCoor(gro_file: str = "", ring_1_id: list = [], ring_2_id: list = []) -> tuple:
    """
    getCoor: to get the coordinates of atoms of two rings

    :parameter:
        gro_file: the gro filename
        ring_1_id: the atom ids of ring 1
        ring_2_id: the atom ids of ring 2

    :return:
        time: a list to store frames
        ring_1_frames: a list to store coordinates of frames
        ring_2_frames: a list to store coordinates of frames
    """

    with open(gro_file, "r") as fo:
        lines = fo.readlines()
    frames = []
    atom_lines = []
    for line in lines:
        ## not the atom line, judged by length of line
        if len(line) != 45 and len(line) != 69:
            # number line
            if len(line.strip().split()) == 1:
                if len(atom_lines) != 0:
                    frames.append(atom_lines)
                atom_lines = []
        elif len(line) == 45 or len(line) == 69:
            # suit for atom number < 10000
            if len(line[20:44].split()) == 3 and len(line[15:44].split()) == 4:
                atom_lines.append(line.rstrip())
    ## add the last frame
    if len(atom_lines) != 0:
        frames.append(atom_lines)
    ## read the atom coor
    ring_1_frames = []
    ring_2_frames = []
    for frame in frames:
        ring_1_coor = []
        ring_2_coor = []
        for line in frame:
            if int(line[15:20]) in ring_1_id:
                ring_1_coor.append(
                    [float(line[20:28]), float(line[28:36]), float(line[36:44])]
                )
            if int(line[15:20]) in ring_2_id:
                ring_2_coor.append(
                    [float(line[20:28]), float(line[28:36]), float(line[36:44])]
                )
        if len(ring_1_coor) != len(ring_1_id) or len(ring_2_coor) != len(ring_2_id):
            logging.error("shit happens when reading coordinates of ring")
            sys.exit()
        ring_1_frames.append(ring_1_coor)
        ring_2_frames.append(ring_2_coor)
    ## new a time sequence
    time = [i for i in range(len(frames))]

    return time, ring_1_frames, ring_2_frames


def dealNdx(ndx_file: str = "", select: list = [], vg: bool = False) -> tuple:
    """
    dealNdx: to process the index file

    :parameter:
        ndx_file: the index filename
        select: a list to store the groupnames specified
        vg: whether to use a group in index file for vector calculation

    :return:
        ring_1_id: a list to store the atom id of ring 1
        ring_2_id: a list to store the atom id of ring 2
    """

    with open(ndx_file, "r") as fo:
        content = fo.read()
    ## read in each group
    ndx_dic = {}
    content = content.strip().strip("[").replace("\n", " ")
    ndx_groups = content.split("[")
    for group in ndx_groups:
        items = group.split("]")
        if items[0].strip() in ndx_dic.keys():
            logging.error("two groups with the same name")
            sys.exit()
        ndx_dic[items[0].strip()] = [
            int(n.strip()) for n in items[1].split() if n != ""
        ]
    if select == None:
        ## print to get input from user
        logging.info("reading your index file:")
        for name, num_lis in ndx_dic.items():
            print("    {:30}   {:8} atoms".format(name, len(num_lis)))
            # print(" ".join([ str(i) for i in num_lis]))
        if vg == False:
            ring_1_name = input("Type the name of first ring -> ")
            logging.info("you have chosed " + ring_1_name + " as first ring.")
            ring_2_name = input("Type the name of second ring -> ")
            logging.info("you have chosed " + ring_2_name + " as second ring.")
        elif vg == True:
            ring_1_name = input("Type the name of ring group -> ")
            logging.info("you have chosed " + ring_1_name + " as first ring.")
            ring_2_name = input("Type the name of vector group-> ")
            logging.info("you have chosed " + ring_2_name + " as vector group.")
    elif len(select) == 2:
        if vg == False:
            ring_1_name = select[0]
            logging.info("you have chosed " + ring_1_name + " as first ring.")
            ring_2_name = select[1]
            logging.info("you have chosed " + ring_2_name + " as second ring.")
        elif vg == True:
            ring_1_name = select[0]
            logging.info("you have chosed " + ring_1_name + " as first ring.")
            ring_2_name = select[1]
            logging.info("you have chosed " + ring_2_name + " as vector group.")
    else:
        logging.error("Wrong parameter number of -select")
        sys.exit()

    ## atom id of two ring groups
    ring_1_id = ndx_dic[ring_1_name]
    ring_2_id = ndx_dic[ring_2_name]

    ## return the atom ids of these two rings
    return ring_1_id, ring_2_id


def dealNdx_single(ndx_file: str = "", select: list = []) -> list:
    """
    dealNdx_single: to deal with index file with one group needed

    :parameter:
        ndx_file: the index filename
        select: a list to store index group name specified in command line

    :return:
        ring_1_id: a list to store the atom id
    """

    with open(ndx_file, "r") as fo:
        content = fo.read()
    ## read in each group
    ndx_dic = {}
    content = content.strip().strip("[").replace("\n", " ")
    ndx_groups = content.split("[")
    for group in ndx_groups:
        items = group.split("]")
        if items[0].strip() in ndx_dic.keys():
            logging.error("two groups with the same name")
            sys.exit()
        ndx_dic[items[0].strip()] = [
            int(n.strip()) for n in items[1].split() if n != ""
        ]
    ## print to get input from user
    logging.info("reading your index file:")
    for name, num_lis in ndx_dic.items():
        print("    {:30}   {:8} atoms".format(name, len(num_lis)))
    # print(" ".join([ str(i) for i in num_lis]))
    if select == None:
        ring_1_name = input("Type the name of first ring -> ")
        logging.info("you have chosed " + ring_1_name + " as first ring.")
    elif len(select) == 1:
        ring_1_name = select[0]
        logging.info("you have chosed " + ring_1_name + " as first ring.")
    else:
        logging.error("you may only need one parameter for select")
        sys.exit()

    ## atom id of two ring groups
    ring_1_id = ndx_dic[ring_1_name]

    ## return the atom ids of these two rings
    return ring_1_id


def calcAng_RingVec(ring_frames: list = [], vec_frames: list = []) -> list:
    """
    calcAng_RingVec: to calculate the anglef between a ring and a vector

    :parameter:
        ring_frames: a list to store atom coordinates of frames
        vec_frames: a list to store vector of frames

    :return:
        angles: a list to store angles
    """

    ## check the frames
    if len(ring_frames) != len(vec_frames):
        logging.error("length of frames of ring and vector isn't equal")
        sys.exit()
    ## calculate the angles
    angles = []
    for i in range(len(ring_frames)):
        ring_1_coor = ring_frames[i]
        vector = vec_frames[i]
        ## p1 = [ a1, a2, a3 ], p2 = [ b1, b2, b3 ]
        ## p1xp2 = [ a2b3 - a3b2, a3b1 - a1b3, a1b2 - a2b1 ]
        ## calculate the plane normal of ring 1 by cross product
        r1_p1 = [
            ring_1_coor[2][0] - ring_1_coor[0][0],
            ring_1_coor[2][1] - ring_1_coor[0][1],
            ring_1_coor[2][2] - ring_1_coor[0][2],
        ]
        r1_p2 = [
            ring_1_coor[4][0] - ring_1_coor[0][0],
            ring_1_coor[4][1] - ring_1_coor[0][1],
            ring_1_coor[4][2] - ring_1_coor[0][2],
        ]
        r1_xProd = [
            r1_p1[1] * r1_p2[2] - r1_p1[2] * r1_p2[1],
            r1_p1[2] * r1_p2[0] - r1_p1[0] * r1_p2[2],
            r1_p1[0] * r1_p2[1] - r1_p1[1] * r1_p2[0],
        ]
        ## calc the degree
        dotProduct = r1_xProd[0] * vector[0] + r1_xProd[1] * vector[1]
        dotProduct += r1_xProd[2] * vector[2]
        r1_xProd_norm = (r1_xProd[0] ** 2 + r1_xProd[1] ** 2 + r1_xProd[2] ** 2) ** 0.5
        vector_norm = (vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2) ** 0.5
        cos_degree = dotProduct / (r1_xProd_norm * vector_norm)
        degree = math.acos(cos_degree) * 180 / math.pi
        if degree > 90:
            degree = 180 - degree
        angles.append(degree)
    return angles


def calcVec(vg_frames: list = []) -> list:
    """
    calcVec: to calculate the vector from vector group

    :parameter:
        vg_frames: a list to store vector atom coordinates of frames

    :return:
        vg_vec_frames: a list to store result vectors
    """

    vg_vec_frames = []
    for atoms in vg_frames:
        vec = [0, 0, 0]
        for i in range(1, len(atoms)):
            vec[0] += atoms[i][0] - atoms[i - 1][0]
            vec[1] += atoms[i][1] - atoms[i - 1][1]
            vec[2] += atoms[i][2] - atoms[i - 1][2]
        vec = [v / (len(atoms) - 1) for v in vec]
        vg_vec_frames.append(vec)
    return vg_vec_frames


def dealTwoRings(
    ndx_file: str = "",
    gro_file: str = "",
    time_b: int = 0,
    time_dt: int = 1,
    output_file: str = "",
    select: list = [],
) -> None:
    """
    dealTwoRings: to calculate the distance and angles of two rings

    :parameter:
        ndx_file: index filename
        gro_file: gro filename
        time_b: the start frame
        time_dt: the frame interval
        output_file: a filename for output
        select: a list to store the selected index group name
    """

    ## get the atom id
    ring_1_id, ring_2_id = dealNdx(ndx_file, select, False)
    if (
        len(ring_1_id) < 5
        or len(ring_1_id) > 7
        or len(ring_2_id) < 5
        or len(ring_2_id) > 7
    ):
        logging.error("index of your ring is more than 7 or less than 5")
        logging.error("only support 5, 6 or 7 membered ring which is in a plane ")
        logging.error("please check your index file")
        logging.error("your index : {} {}".format(ring_1_id, ring_2_id))
        sys.exit()
    ## get the coordinates of two rings
    time, ring_1_frames, ring_2_frames = getCoor(gro_file, ring_1_id, ring_2_id)
    ## modify the time sequence
    time = [t * time_dt + time_b for t in time]
    # print(len(time), len(ring_1_frames), len(ring_2_frames))
    ## calculate the distance of two rings
    distance = calcDist(ring_1_frames, ring_2_frames)
    ## calculate the angles of the normals of two rings
    angles = calcAng(ring_1_frames, ring_2_frames)
    # print(len(time), len(distance), len(angles))
    ## check data and output
    logging.info("there is {} frames in your gro file".format(len(time)))
    if len(time) != len(distance) or len(time) != len(angles):
        logging.error("length of time, dist, ang are not equal")
        logging.info(
            "time: {}; distance: {}; angles: {}".format(
                len(time), len(distance), len(angles)
            )
        )
        sys.exit()
    out_content = "# This file was created by DIT pipi_dist_ang\n"
    out_content += """@    title "Dist_Ang"\n"""
    out_content += """@    xaxis  label "Frames"\n"""
    out_content += """@    yaxis  label ""\n"""
    out_content += """@TYPE xy\n@ view 0.15, 0.15, 0.75, 0.85\n"""
    out_content += """@ legend on\n@ legend box on\n@ legend loctype view\n"""
    out_content += """@ legend 0.78, 0.8\n@ legend length 2\n"""
    out_content += """@ s0 legend "Dist (nm)"\n@ s1 legend "Angle"\n"""
    out_content += "\n".join(
        [
            "{:<10.0f} {:>10.3f} {:>10.3f} ".format(time[i], distance[i], angles[i])
            for i in range(len(distance))
        ]
    )
    with open(output_file, "w") as fo:
        fo.write(out_content)

    ## calc the angle distribution
    ang0_30, ang30_60, ang60_90 = 0, 0, 0
    for ang in angles:
        if ang >= 0 and ang < 30:
            ang0_30 += 1
        elif ang >= 30 and ang < 60:
            ang30_60 += 1
        elif ang > 60:
            ang60_90 += 1
    print("-" * 79)
    print(
        " =>  0 <= angle < 30 : {}/{} = {:>6.2%}".format(
            ang0_30, len(time), ang0_30 * 1.0 / len(time)
        )
    )
    print(
        " => 30 <= angle < 60 : {}/{} = {:>6.2%}".format(
            ang30_60, len(time), ang30_60 * 1.0 / len(time)
        )
    )
    print(
        " => 60 <= angle < 90 : {}/{} = {:>6.2%}".format(
            ang60_90, len(time), ang60_90 * 1.0 / len(time)
        )
    )
    print("-" * 79)
    ## calc the average distance
    logging.info("average distance : {:>10.4f} nm".format(sum(distance) / len(time)))


def dealRingVG(
    ndx_file: str = "",
    gro_file: str = "",
    time_b: int = 0,
    time_dt: int = 1,
    output_file: str = "",
    select: list = [],
) -> None:
    """
    dealRingVG: to calculate the distance and angles of ring and vector from index group

    :parameter:
        ndx_file: index filename
        gro_file: gro filename
        time_b: the start frame
        time_dt: the frame interval
        output_file: a filename for output
        select: a list to store the selected index group name
    """

    ring_id, vg_id = dealNdx(ndx_file, select, True)
    if len(ring_id) < 5 or len(ring_id) > 7:
        logging.error("index of your ring is more than 7 or less than 5")
        logging.error("only support 5, 6 or 7 membered ring which is in a plane ")
        logging.error("please check your index file")
        logging.error("your index : {}".format(ring_id))
        sys.exit()
    if len(vg_id) < 2:
        logging.error("less than 2 atom index in vector group you input")
        logging.error("2 or more atom index are needed")
        logging.error("check your vector index : {}".format(vg_id))
        sys.exit()
    ## get the coordinates of ring and vg
    time, ring_frames, vg_frames = getCoor(gro_file, ring_id, vg_id)
    ## modify the time sequence
    time = [t * time_dt + time_b for t in time]
    # calculate the vector vs time
    vg_vec_frames = calcVec(vg_frames)
    ## calculate the angles
    angles = calcAng_RingVec(ring_frames, vg_vec_frames)
    # save results
    logging.info("there is {} frames in your gro file".format(len(time)))
    if len(time) != len(angles):
        logging.error("length of time, angles are not equal")
        logging.info("time: {}; angles: {}".format(len(time), len(angles)))
    out_content = "# This file was created by DIT pipi_dist_ang\n"
    out_content += """@    title "Angle of ring and vector"\n"""
    out_content += """@    xaxis  label "Frames"\n"""
    out_content += """@    yaxis  label "Angle"\n"""
    out_content += """@TYPE xy\n@ view 0.15, 0.15, 0.75, 0.85\n"""
    out_content += """@ legend on\n@ legend box on\n@ legend loctype view\n"""
    out_content += """@ legend 0.78, 0.8\n@ legend length 2\n"""
    out_content += "\n".join(
        ["{:<10.0f} {:>10.3f} ".format(time[i], angles[i]) for i in range(len(angles))]
    )
    with open(output_file, "w") as fo:
        fo.write(out_content)
    ## calc the angle distribution
    ang0_30, ang30_60, ang60_90 = 0, 0, 0
    for ang in angles:
        if ang >= 0 and ang < 30:
            ang0_30 += 1
        elif ang >= 30 and ang < 60:
            ang30_60 += 1
        elif ang > 60:
            ang60_90 += 1
    print("-" * 79)
    print(
        " =>  0 <= angle < 30 : {}/{} = {:>6.2%}".format(
            ang0_30, len(time), ang0_30 * 1.0 / len(time)
        )
    )
    print(
        " => 30 <= angle < 60 : {}/{} = {:>6.2%}".format(
            ang30_60, len(time), ang30_60 * 1.0 / len(time)
        )
    )
    print(
        " => 60 <= angle < 90 : {}/{} = {:>6.2%}".format(
            ang60_90, len(time), ang60_90 * 1.0 / len(time)
        )
    )
    print("-" * 79)


def dealRingVec(
    ndx_file: str = "",
    gro_file: str = "",
    time_b: int = 0,
    time_dt: int = 1,
    output_file: str = "",
    vec: list = [],
    select: list = [],
) -> None:
    """
    dealRingVec: to calculate the distance and angles of ring and vector

    :parameter:
        ndx_file: index filename
        gro_file: gro filename
        time_b: the start frame
        time_dt: the frame interval
        output_file: a filename for output
        vec: a list to store vector
        select: a list to store the selected index group name
    """

    ring_id = dealNdx_single(ndx_file, select)
    if len(ring_id) < 5 or len(ring_id) > 7:
        logging.error("index of your ring is more than 7 or less than 5")
        logging.error("only support 5, 6 or 7 membered ring which is in a plane ")
        logging.error("please check your index file")
        logging.error("your index : {}".format(ring_id))
        sys.exit()
    time, ring_frames, _ = getCoor(gro_file, ring_id, ring_id)
    ## modify the time sequence
    time = [t * time_dt + time_b for t in time]
    # deal with vec
    vec = [float(i) for i in vec]
    if 0 == vec[0] == vec[1] == vec[2]:
        logging.error("You can't input an all zero vector")
        sys.exit()
    vec_frames = [vec for i in range(len(time))]
    # calculate the angles
    angles = calcAng_RingVec(ring_frames, vec_frames)
    # save results
    logging.info("there is {} frames in your gro file".format(len(time)))
    if len(time) != len(angles):
        logging.error("length of time, angles are not equal")
        logging.info("time: {}; angles: {}".format(len(time), len(angles)))
    out_content = "# This file was created by DIT pipi_dist_ang\n"
    out_content += """@    title "Angle of ring and vector"\n"""
    out_content += """@    xaxis  label "Frames"\n"""
    out_content += """@    yaxis  label "Angle"\n"""
    out_content += """@TYPE xy\n@ view 0.15, 0.15, 0.75, 0.85\n"""
    out_content += """@ legend on\n@ legend box on\n@ legend loctype view\n"""
    out_content += """@ legend 0.78, 0.8\n@ legend length 2\n"""
    out_content += "\n".join(
        ["{:<10.0f} {:>10.3f} ".format(time[i], angles[i]) for i in range(len(angles))]
    )
    with open(output_file, "w") as fo:
        fo.write(out_content)
    ## calc the angle distribution
    ang0_30, ang30_60, ang60_90 = 0, 0, 0
    for ang in angles:
        if ang >= 0 and ang < 30:
            ang0_30 += 1
        elif ang >= 30 and ang < 60:
            ang30_60 += 1
        elif ang > 60:
            ang60_90 += 1
    print("-" * 79)
    print(
        " =>  0 <= angle < 30 : {}/{} = {:>6.2%}".format(
            ang0_30, len(time), ang0_30 * 1.0 / len(time)
        )
    )
    print(
        " => 30 <= angle < 60 : {}/{} = {:>6.2%}".format(
            ang30_60, len(time), ang30_60 * 1.0 / len(time)
        )
    )
    print(
        " => 60 <= angle < 90 : {}/{} = {:>6.2%}".format(
            ang60_90, len(time), ang60_90 * 1.0 / len(time)
        )
    )
    print("-" * 79)


def pipi_dist_ang(
    ndx_file: str = "",
    gro_file: str = "",
    time_b: int = 0,
    time_dt: int = 1,
    output_file: str = "",
    vg: bool = False,
    vec: list = [],
    select: list = [],
) -> None:
    """
    pipi_dist_ang: a list to check parameters and start init calculation

    :parameter:
        ndx_file: index filename
        gro_file: gro filename
        time_b: the start frame
        time_dt: the frame interval
        output_file: a filename for output
        vg: whether to get vector from atom id in index file
        vec: a list to store vector
        select: a list to store the selected index group name
    """

    if ndx_file == "":
        logging.error("please specify index file by -n")
        sys.exit()
    if gro_file == "":
        logging.error("please specify gro file by -f")
        sys.exit()
    if output_file == "":
        logging.error("please specify output file by -o")
        sys.exit()
    if not os.path.exists(ndx_file):
        logging.error("no {} in current directory".format(ndx_file))
        sys.exit()
    if not os.path.exists(gro_file):
        logging.error("no {} in current directory".format(gro_file))
        sys.exit()
    if os.path.exists(output_file):
        logging.error("already one {} in current directory".format(output_file))
        sys.exit()

    if vec == None and vg == False:
        dealTwoRings(ndx_file, gro_file, time_b, time_dt, output_file, select)
    elif vec == None and vg == True:
        dealRingVG(ndx_file, gro_file, time_b, time_dt, output_file, select)
    elif vec != None and vg == False:
        dealRingVec(ndx_file, gro_file, time_b, time_dt, output_file, vec, select)
    elif vec != None and vg == True:
        logging.error("You can't set -vg and -vec at the same time")
        sys.exit()


def pipi_dist_ang_call_functions(arguments: list = []):
    """call functions according to arguments"""

    if len(arguments) == 0:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    description = "To calculate the distance and angles of two 5 or 6 membered"
    description += "rings or vectors you defined"
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--input", help="gro file")
    parser.add_argument("-n", "--index", help="index file contains two groups of rings")
    parser.add_argument("-b", default=0, type=int, help="set the start time, default=0")
    parser.add_argument(
        "-dt", default=1, type=int, help="set the time interval, default=1"
    )
    parser.add_argument(
        "-o",
        "--output",
        default="output.xvg",
        help="the results data, default output.xvg",
    )
    parser.add_argument(
        "-vg", action="store_true", help="whether to get vector by index group"
    )
    parser.add_argument(
        "-vec", nargs=3, help="get vector by your input, eg. -vec 6 6 6"
    )
    parser.add_argument(
        "-select", nargs="*", help="select the groups, eg. -select ring1 ring2"
    )

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        sys.exit()

    method = arguments[1]
    # print(method)
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        sys.exit()

    if len(arguments) == 2:
        logging.error("no parameters, type 'dit pip_dist_ang -h' for more infos.")
        sys.exit()
    args = parser.parse_args(arguments[2:])
    if method == "pipi_dist_ang":
        pipi_dist_ang(
            args.index,
            args.input,
            args.b,
            args.dt,
            args.output,
            args.vg,
            args.vec,
            args.select,
        )
    else:
        logging.error("unknown method {}".format(method))
        sys.exit()

    logging.info("good day !")


def main():
    pipi_dist_ang_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
DSSP module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to deal and present figures about secondary structure

This DSSP module contains:
    - DSSP function

This file is provided to you by GPLv3 license."""


import os
import sys
import logging
import argparse
import numpy as np
import matplotlib.pyplot as plt


base = os.path.dirname(os.path.realpath(__file__))
if base not in sys.path:
    sys.path.insert(0, base)

from XPM import XPM


def dssp(
    xpmfile: str = None,
    xlabel: str = None,
    ylabel: str = None,
    title: str = None,
    xshrink: float = 1.0,
):
    """read xpm file, generate xpm figure, stacked Residue vs time, stacked time occupancy vs residue"""

    xpm = XPM(xpmfile, xlabel, ylabel, title, xshrink)
    xpm.draw_origin(False, None, False)

    xpm.xpm_datalines.reverse()
    xpm.xpm_yaxis.reverse()
    xpm_datamatrix = []
    for dataline in xpm.xpm_datalines:
        dot_list = []
        for i in range(
            0, xpm.xpm_width * xpm.xpm_char_per_pixel, xpm.xpm_char_per_pixel
        ):
            dot_list.append(xpm.chars.index(dataline[i : i + xpm.xpm_char_per_pixel]))
        xpm_datamatrix.append(dot_list)

    # residue_occupancy
    residue_occupancy_dic = {k: [] for k in xpm.notes}
    for x, xaxis in enumerate(xpm.xpm_xaxis):
        for key in residue_occupancy_dic.keys():
            residue_occupancy_dic[key].append(0)
        for y, yaxis in enumerate(xpm.xpm_yaxis):
            dot = xpm_datamatrix[y][x]
            key = xpm.notes[dot]
            residue_occupancy_dic[key][-1] += 1
    labels = [k for k in residue_occupancy_dic.keys()]
    labels.reverse()
    data_columns = [residue_occupancy_dic[l] for l in labels]
    if "residue_occupancy_data.csv" in os.listdir():
        logging.error(
            "residue_occupancy_data.csv already in current directory, unable to output file with same name"
        )
        sys.exit()
    with open("residue_occupancy_data.csv", "w") as fo:
        fo.write(",".join(labels) + "\n")
        for t in range(len(data_columns[0])):
            fo.write(
                ",".join([str(data_columns[d][t]) for d in range(len(data_columns))])
                + "\n"
            )
    ylim_max, ylim_min = 0, 0
    for index, _ in enumerate(labels):
        stack_data = [
            sum([data_columns[c][x] for c in range(index, len(labels))])
            for x in range(len(xpm.xpm_xaxis))
        ]
        plt.fill_between(
            xpm.xpm_xaxis,
            stack_data,
            [0 for _ in range(len(stack_data))],
            label=labels[index],
        )
        ylim_max = (ylim_max, max(stack_data))[ylim_max < max(stack_data)]
        ylim_min = (ylim_min, min(stack_data))[ylim_min > min(stack_data)]

    plt.xlabel(xpm.xpm_xlabel)
    plt.ylabel(xpm.xpm_ylabel)
    plt.title("Stacked plot of " + xpm.xpm_title)
    plt.xlim(
        np.min(xpm.xpm_xaxis),
        np.max(xpm.xpm_xaxis),
    )
    plt.ylim(ylim_min, ylim_max)
    plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()
    plt.show()

    # time_occupancy
    time_occupancy_dic = {k: [] for k in xpm.notes}
    for y, yaxis in enumerate(xpm.xpm_yaxis):
        for key in time_occupancy_dic.keys():
            time_occupancy_dic[key].append(0)
        for x, xaxis in enumerate(xpm.xpm_xaxis):
            dot = xpm_datamatrix[y][x]
            key = xpm.notes[dot]
            time_occupancy_dic[key][-1] += 1
    labels = [k for k in time_occupancy_dic.keys()]
    labels.reverse()
    data_columns = [time_occupancy_dic[l] for l in labels]
    if "time_occupancy_data.csv" in os.listdir():
        logging.error(
            "time_occupancy_data.csv already in current directory, unable to output file with same name"
        )
        sys.exit()
    with open("time_occupancy_data.csv", "w") as fo:
        fo.write(",".join(labels) + "\n")
        for t in range(len(data_columns[0])):
            fo.write(
                ",".join([str(data_columns[d][t]) for d in range(len(data_columns))])
                + "\n"
            )
    ylim_max, ylim_min = 0, 0
    for index, _ in enumerate(labels):
        stack_data = [
            sum([data_columns[c][x] for c in range(index, len(labels))])
            for x in range(len(xpm.xpm_yaxis))
        ]
        plt.bar(xpm.xpm_yaxis, stack_data, label=labels[index], width=1)
        # plt.fill_between(
        #     xpm.xpm_yaxis,
        #     stack_data,
        #     [0 for _ in range(len(stack_data))],
        #     label=labels[index],
        # )
        ylim_max = (ylim_max, max(stack_data))[ylim_max < max(stack_data)]
        ylim_min = (ylim_min, min(stack_data))[ylim_min > min(stack_data)]

    plt.xlabel(xpm.xpm_ylabel)
    plt.ylabel(xpm.xpm_xlabel)
    ytics, _ = plt.yticks()
    yticklabels= [xpm.xpm_xaxis[int(y)] for y in ytics[:-1]]
    plt.yticks(ytics[:-1], yticklabels)
    plt.title("Stacked plot of " + xpm.xpm_title)
    plt.xlim(np.min(xpm.xpm_yaxis) - 0.5, np.max(xpm.xpm_yaxis) + 0.5)
    plt.ylim(ylim_min, ylim_max)
    plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()
    plt.show()


def dssp_call_functions(arguments: list = []):
    """call functions according to arguments"""

    if len(arguments) == 0:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(
        description="hbond command process the hbond infos"
    )
    parser.add_argument("-f", "--input", help="the xpm file for input")
    parser.add_argument(
        "-xs",
        "--xshrink",
        type=float,
        default=1.0,
        help="modify x-axis by multipling xshrink",
    )
    parser.add_argument("-x", "--xlabel", type=str, help="the xlabel of figure")
    parser.add_argument("-y", "--ylabel", type=str, help="the ylabel of figure")
    parser.add_argument("-t", "--title", type=str, help="the title of figure")

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        exit()

    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        exit()
    if len(arguments) == 2:
        logging.error("no parameters, type 'dit <command> -h' for more infos.")
        exit()

    args = parser.parse_args(arguments[2:])

    if method == "dssp":
        xpmfile = args.input
        xshrink = args.xshrink
        dssp(xpmfile, args.xlabel, args.ylabel, args.title, xshrink)
    else:
        logging.error("unknown method {}".format(method))
        exit()

    logging.info("May you good day !")


def main():
    dssp_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
ndxParser module is part of DuIvyTools for parsing the ndx file generated by GROMACS.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import math
from typing import List, Union

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log


class NDX(log):
    """NDX class was designed to parse index file"""

    def __init__(self, ndxfile: str, new_file: bool = False) -> None:
        self.ndxfile: str = ndxfile
        self.indexs: List[int] = []
        self.names: List[str] = []
        self.column_nums: List[int] = []

        if not new_file and ndxfile:
            if not os.path.exists(ndxfile):
                self.error(f"No {ndxfile} detected! check it!")
            if ndxfile[-4:] != ".ndx":
                self.error(
                    f"you must specify a file with suffix .ndx, instead of {ndxfile}"
                )
            with open(ndxfile, "r") as fo:
                lines = [line.strip() for line in fo.readlines()]
            for id, line in enumerate(lines):
                if line.strip() == "":
                    continue
                elif line[0] == "[" and line[-1] == "]":
                    name = line[1:-1].strip()
                    self.indexs.append([])
                    self.column_nums.append(0)
                    self.names.append(name)
                else:
                    try:
                        line_list = [int(i) for i in line.split()]
                        if len(self.indexs[-1]) == 0:
                            self.column_nums[-1] = len(line_list)
                        self.indexs[-1] += line_list
                    except:
                        self.error(f"Unable to parse line {id} of {ndxfile}, check it!")
        if len(self.names) != len(self.indexs) or len(self.names) != len(
            self.column_nums
        ):
            self.critical("wrong length in paring ndx file")

    def __len__(self) -> int:
        """return number of groups"""
        return len(self.names)

    def __getitem__(
        self, key: Union[str, int]
    ) -> Union[Union[str, None], Union[None, List[int]]]:
        """get item by group id or group name"""
        if isinstance(key, int):
            if key < len(self):
                name = self.names[key]
                index = self.indexs[key]
                return name, index
            else:  # index over range
                return None, None
        elif isinstance(key, str):
            if key in self.names:
                index = self.indexs[self.names.index(key)]
                return key, index
            else:
                return None, None
        else:
            return None, None

    def __setitem__(self, key: Union[str, int], indexs: List[int]) -> None:
        """set item by group id of group name"""
        if isinstance(key, int):
            if key >= len(self):
                self.error("key over range for setting item of NDX")
            self.names[id] = key
            self.indexs[id] = indexs
            self.column_nums[id] = 15
        elif isinstance(key, str):
            if key not in self.names:
                self.indexs.append(indexs)
                self.names.append(key)
                self.column_nums.append(15)
            else:
                ids = self.get_id_by_name()
                for id in ids:
                    self.names[id] = key
                    self.indexs[id] = indexs
                    self.column_nums[id] = 15
        else:
            pass

    def add(self, name: str, indexs: List[int], column_num: int = 15) -> None:
        """add a new group to NDX"""
        self.names.append(name)
        self.indexs.append(indexs)
        self.column_nums.append(column_num)

    def __delitem__(self, key: Union[str, int]) -> None:
        """delete group by group id or group name"""
        if isinstance(key, int):
            if key >= len(self):
                self.error("key over range for deleting item of NDX")
            del self.names[key]
            del self.indexs[key]
            del self.column_nums[key]
        elif isinstance(key, str):
            name = key
            while name in self.names:
                id = self.names.index(name)
                del self.names[id]
                del self.indexs[id]
                del self.column_nums[id]
        else:
            self.error("Wrong key for deleting item of NDX")

    def get_id_by_name(self, key: str) -> List[int]:
        """get group id by group name"""
        id_list = []
        for id, name in enumerate(self.names):
            if key == name:
                id_list.append(id)
        return id_list

    @property
    def show_names(self) -> str:
        """convert group names and infos into string and return"""
        output: str = ""
        for id, name in enumerate(self.names):
            output += f"{id} {name:<20} {len(self.indexs[id])}\n"
        return output

    def save(self, outfile) -> None:
        """dump NDX class into index file"""
        with open(outfile, "w") as fo:
            fo.write(str(self))

    def __str__(self) -> str:
        """convert NDX class into string"""
        output: str = ""
        for id in range(len(self)):
            out = self.formatter(id)
            output += out
        output += "\n "
        return output

    def formatter(self, id: int) -> str:
        """format group into string by group id"""
        name = self.names[id]
        index = self.indexs[id]
        column_num = self.column_nums[id]
        if column_num <= 0:
            self.error(f"Unable to format indexs by 0 column at group id {id}")
        output: str = f"[ {name} ] \n"
        count = math.ceil(len(index) / column_num)
        for c in range(count):
            items = [f"{v:>4d}" for v in index[c * column_num : (c + 1) * column_num]]
            output += " ".join(items) + "\n"
        return output



"""
xvgParser module is part of DuIvyTools for parsing the xvg file generated by GROMACS.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import time
from typing import List, Tuple, Union

import numpy as np
import scipy.stats as stats

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log


class XVG(log):
    """XVG class for parsing xvg file"""

    def __init__(
        self,
        xvgfile: Union[str, List[str]],
        is_file: bool = True,
        new_file: bool = False,
    ) -> None:
        self.comments: str = ""
        self.comments_tail: str = ""
        self.title: str = ""
        self.xlabel: str = ""
        self.ylabel: str = ""
        self.xmin: float = None
        self.xmax: float = None
        self.ymin: float = None
        self.ymax: float = None
        self.legends: List[str] = []
        self.column_num: int = 0
        self.row_num: int = 0
        self.data_heads: List[str] = []
        self.data_columns: List[Union(float, str)] = []

        if new_file:
            self.xvgfile = xvgfile
        else:
            if is_file:
                self.xvgfile: str = xvgfile
                if not os.path.exists(xvgfile):
                    self.error(f"No {xvgfile} detected ! check it !")
                if xvgfile[-4:] != ".xvg":
                    self.error(
                        f"you must specify a file with suffix .xvg, instead of {xvgfile}"
                    )
                with open(xvgfile, "r") as fo:
                    lines = fo.readlines()
            else:
                lines = xvgfile
            self.parse_xvg(lines)
            if is_file:
                self.info(f"parsing data from {xvgfile} successfully !")

    def parse_xvg(self, lines: List[str]) -> None:
        """parse xvg content (lines) into XVG class

        Args:
            lines (List[str]): xvg file lines
        """
        ## parse data
        for line in lines:
            line = line.strip()
            if line.startswith("#") or line.startswith("&"):
                self.comments += line + "\n"
            elif line.startswith("@"):
                if " title " in line:
                    self.title = line.strip('"').split('"')[-1]
                elif " xaxis " in line and " label " in line:
                    self.xlabel = line.strip('"').split('"')[-1]
                elif " yaxis " in line and " label " in line:
                    self.ylabel = line.strip('"').split('"')[-1]
                elif line.startswith("@ s") and " legend " in line:
                    self.legends.append(line.strip('"').split('"')[-1])
                elif " world xmin " in line:
                    self.xmin = float(line.split()[-1])
                elif " world xmax " in line:
                    self.xmax = float(line.split()[-1])
                elif " world ymin " in line:
                    self.ymin = float(line.split()[-1])
                elif " world ymax " in line:
                    self.ymax = float(line.split()[-1])
                else:
                    pass
            else:
                items = line.split()
                if self.column_num == 0:
                    self.data_columns = [[] for _ in range(len(items))]
                    self.column_num = len(items)
                    self.row_num = 0
                if len(items) != self.column_num:
                    self.error(
                        f"the number of columns in {self.xvgfile} is not equal at line {self.row_num}"
                    )
                for i in range(len(items)):
                    self.data_columns[i].append(items[i])
                self.row_num += 1

        ## convert data
        self.data_heads.append(self.xlabel)
        self.data_columns[0] = [float(c) for c in self.data_columns[0]]
        if len(self.legends) == 0 and len(self.data_columns) > 1:
            self.data_heads.append(self.ylabel)
            self.data_columns[1] = [float(c) for c in self.data_columns[1]]

        if len(self.legends) > 0 and self.column_num > len(self.legends):
            items = [item.strip() for item in self.ylabel.split(",")]
            heads = [l for l in self.legends]
            if len(items) == len(self.legends):
                for i in range(len(items)):
                    heads[i] += " " + items[i]
            elif (
                len(items) == 1
                and items[0] != ""
                and (items[0][0] == "(" and items[0][-1] == ")")
            ):
                for i in range(len(heads)):
                    heads[i] += " " + items[0]
            else:
                self.warn("failed to pair ylabel to legends, use legends in xvg file")
            self.data_heads += heads
            for i in range(len(heads)):
                self.data_columns[i + 1] = [float(c) for c in self.data_columns[i + 1]]

        ## check infos
        for c in range(self.column_num):
            if len(self.data_columns[c]) != self.row_num:
                self.error(f"length of column {c} is not equal to other columns")
        if self.column_num == 0 or self.row_num == 0:
            self.error(f"no data line detected in {self.xvgfile}")
        if len(self.data_heads) < self.column_num:
            self.warn(
                f"string column may detected, data_heads {len(self.data_heads)} < column_num {self.column_num}"
            )

    def save(self, outxvg: str, check: bool = True) -> None:
        """dump XVG class to xvg file

        Args:
            outxvg (str): output xvg file name
            check (bool): whether check XVG attributes before dumping
        """
        if check:
            if len(self.data_heads) == 0:
                self.error("unable to dump xvg with empty data_heads")
            if len(self.data_heads) != 0 and len(self.legends) == 0:
                self.legends = self.data_heads[1:]
            if len(self.data_heads) != 0 and not self.xlabel:
                self.xlabel = self.data_heads[0]
            if len(self.data_columns) == 0 or len(self.data_columns[0]) == 0:
                self.error("unable to dump with empty data_columns")
            if len(self.data_columns) != 0 and self.column_num == 0:
                self.column_num = len(self.data_columns)
            if len(self.data_columns[0]) != 0 and self.row_num == 0:
                self.row_num = len(self.data_columns[0])

        time_info = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        outstr: str = f"# This file was created by DuIvyTools at {time_info}\n"
        if self.comments:
            outstr += self.comments.strip() + "\n"
        outstr += f'@    title "{self.title}"\n'
        if self.xlabel:
            outstr += f'@    xaxis label "{self.xlabel}"\n'
        if self.ylabel:
            outstr += f'@    yaxis label "{self.ylabel}"\n'
        outstr += "@TYPE xy\n@ view 0.15, 0.15, 0.75, 0.85\n"
        outstr += "@ legend on\n@ legend box on\n@ legend loctype view\n"
        outstr += f"@ legend 0.78, 0.8\n@ legend length {len(self.legends)}\n"
        for i, leg in enumerate(self.legends):
            outstr += f'@ s{i} legend "{leg}"\n'
        for row in range(self.row_num):
            for i in range(self.column_num):
                value = self.data_columns[i][row]
                if isinstance(value, int):
                    outstr += f"{value:>6d} "
                else:
                    outstr += f"{value:>16.6f} "
            outstr += "\n"
        if self.comments_tail:
            outstr += self.comments_tail.strip() + "\n"
        with open(outxvg, "w") as fo:
            fo.write(outstr)
        self.info(f"dump xvg to {outxvg} successfully")

    def calc_mvave(
        self, windowsize: int, confidence: float, column_index: int
    ) -> Tuple[List]:
        """
        calculate the moving average of each column

        :parameters:
            windowsize: the window size for calculating moving average
            confidence: the confidence to calculate interval
            conlumn_index: the index for the column to calculate

        :return:
            mvaves: a list contains moving average
            highs: the high value of interval of moving averages
            lows: the low value of interval of moving averages
        """

        if windowsize <= 0 or windowsize > int(self.row_num / 2):
            self.error("windowsize value is not proper")
        if confidence <= 0 or confidence >= 1:
            self.error("confidence value is not proper, it should be in (0,1)")

        if column_index < 0 or column_index >= self.column_num:
            self.error("wrong selection of column_index to calculate moving averages")
        column_data = self.data_columns[column_index]
        mvaves = [np.nan for _ in range(windowsize)]
        highs = [np.nan for _ in range(windowsize)]
        lows = [np.nan for _ in range(windowsize)]
        for i in range(windowsize, self.row_num):
            window_data = np.array(column_data[i - windowsize : i])
            ave = np.mean(window_data)
            std = np.std(window_data)
            interval = stats.norm.interval(confidence, ave, std)
            mvaves.append(ave)
            lows.append(interval[0])
            highs.append(interval[1])
        return mvaves, highs, lows

    def calc_ave(
        self, begin: int, end: int, dt: int, column_index: int
    ) -> Tuple[float]:
        """calculate the average of selected column

        Args:
            begin (int): the begin index
            end (int): the end index
            dt (int): the index step
            column_index (int): the selected column index

        Returns:
            Tuple[float]: legend, average, std.err
        """
        if (begin != None and end != None) and (begin >= end):
            self.error("start index should be less than end index")
        if (begin != None and begin >= self.row_num) or (
            end != None and end >= self.row_num
        ):
            self.error(
                f"start or end index should be less than the number of rows {self.row_num} in xvg file"
            )
        if column_index >= self.column_num:
            self.error(
                f"column index selected should be less than column number {self.column_num}"
            )

        column = self.data_columns[column_index]
        legend = self.data_heads[column_index]
        ave = np.average(column[begin:end:dt])
        std = np.std(column[begin:end:dt], ddof=1)
        return legend, ave, std

    def check_column_index(self, column_index: Union[int, List]) -> None:
        """check user-input column index in or not in xpm column range

        Args:
            column_index (Union[int, List]): the user-input column index(s)
        """
        if isinstance(column_index, int):
            column_index = [column_index]
        elif isinstance(column_index, list):
            pass
        else:
            self.error(
                "wrong type specification of column_index for checking the column index range"
            )
        for column in column_index:
            if column >= self.column_num:
                self.error(
                    f"""{self.xvgfile} only contianed {self.column_num} columns. Column index start from 0. You can't select column {column}"""
                )


class XVGS(log):
    """XVGS class for parsing xvg file with multiframes"""

    def __init__(self, xvgfile: str) -> None:
        self.xvgfile: str = xvgfile
        self.frames: list[XVG] = []

        with open(xvgfile, "r") as fo:
            lines = fo.readlines()
        contents: List[List[str]] = [[]]
        for line in lines:
            item = line.strip()
            if item != "&" and item != "":
                contents[-1].append(line)
            elif item == "&":
                contents.append([])

        for content in contents:
            if len(content) == 0:
                continue
            xvg = XVG(content, is_file=False)
            self.frames.append(xvg)
        self.info(f"parsing multi-frames data from {xvgfile} successfully !")

    def __len__(self) -> int:
        """return the frame number"""
        return len(self.frames)

    def __getitem__(self, index: int) -> XVG:
        """get one frame by frame index"""
        return self.frames[index]



"""
MolMap module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to map two pdb files which contain two small molecules
with exactly same comformation.

This module contains:
    - mol_map function

This file is provided to you by GPLv3 license."""


import os
import sys
import argparse
import logging


def similarity(li_0, li_1):
    if len(li_0) != len(li_1):
        print("Wrong")
        exit()
    simi = 0
    for i in range(len(li_0)):
        if li_0[i] != li_1[i]:
            simi += (li_0[i] - li_1[i], li_1[i] - li_0[i])[li_0[i] < li_1[i]]
    return simi


def readpdb2dic(file):
    origin = []
    with open(file, "r") as fo:
        lines = [
            line
            for line in fo.readlines()
            if line.startswith("ATOM") or line.startswith("HETATM")
        ]
    for line in lines:
        atom_num = int(line[6:11])
        atom_name = line[12:16].strip()
        coor_x = float(line[30:38])
        coor_y = float(line[38:46])
        coor_z = float(line[46:54])
        symbol = line[76:78].strip()
        dic = {
            "n": atom_num,
            "a": atom_name,
            "x": coor_x,
            "y": coor_y,
            "z": coor_z,
            "s": symbol,
        }
        origin.append(dic)

    heavy = [dic for dic in origin if dic["s"] != "H"]
    hydrogen = [dic for dic in origin if dic["s"] == "H"]

    ## deal with non-H
    distances = []
    for c1 in heavy:
        dist = []
        for c2 in heavy:
            d = (
                (c1["x"] - c2["x"]) ** 2
                + (c1["y"] - c2["y"]) ** 2
                + (c1["z"] - c2["z"]) ** 2
            ) ** 0.5
            dist.append(round(d * 3))
        distances.append(sorted(dist))

    ## deal with H
    nearest = []
    for c1 in hydrogen:
        dist = []
        for c2 in heavy:
            d = (
                (c1["x"] - c2["x"]) ** 2
                + (c1["y"] - c2["y"]) ** 2
                + (c1["z"] - c2["z"]) ** 2
            ) ** 0.5
            dist.append(d)
        nearest.append(dist.index(min(dist)))

    return origin, heavy, hydrogen, distances, nearest


def mol_map(name_file: str = "", coor_file: str = "", out_file: str = "") -> None:
    """
    mol_map: a function to map atom coordinates of one pdb file to another
    pdb file. The molecule in these two file should share exactly same
    comformation.

    :parameters:
        name_file: the pdb file which provedi infos except coordinates.
        coor_file: the pdb file which provide coordinates.
        out_file: the pdb file for output results.

    :return:
        None
    """

    #####################################################################
    #### 使用注意：输入的两个pdb文件中的分子必须一模一样，构象也必须一样
    ####
    #### 匹配算法说明
    #### 在pdb文件中，只有最后一列的原子类型可以作为参考，因为不能直接根
    #### 据原子名称等信息进行匹配。基于图同构之类的算法又太复杂了，这里
    #### 使用的是一种更简单的思路。
    #### 对于重原子：
    #### 每一个原子与其它所有重原子的距离的列表，应该是唯一的。因而计算
    #### 该原子与其它重原子的距离并从小到大排序，得到该原子的指纹。只需
    #### 要匹配到另一个pdb中相似度最高的指纹的原子，就应当是此原子了
    #### 对于氢原子：
    #### 确定了重原子的匹配之后，氢原子可以通过距离最近的重原子进行匹配，
    #### 对于同一个（匹配的）重原子，一定是唯一的那个氢原子与之距离最近，
    #### 对于等效H，如-CH2等，因为H是等效的，所以只需要匹配了之后剔除已
    #### 匹配的就好了，就算匹配交叉了，也不影响结果的正确性
    #####################################################################

    logging.info("The two molecules you input MUST be with the SAME conformation")

    for file in [name_file, coor_file]:
        if not os.path.exists(file):
            logging.error("no {} in current dirrectory. ".format(file))
            sys.exit()
        if file[-4:] != ".pdb":
            logging.error("mol_map only accept pdb file with suffix .pdb")
            sys.exit()

    origin1, heavy1, hydrogen1, distances1, nearest1 = readpdb2dic(name_file)
    origin2, heavy2, hydrogen2, distances2, nearest2 = readpdb2dic(coor_file)

    maplist = []
    for dli1 in distances1:
        sims = []
        for dli2 in distances2:
            sims.append(similarity(dli1, dli2))
        maplist.append(sims.index(min(sims)))
    Hmaplist = [-1 for _ in nearest1]
    for i1, near1 in enumerate(nearest1):
        for i2, near2 in enumerate(nearest2):
            if maplist[near1] == near2:
                Hmaplist[i1] = i2
                nearest2[i2] = -1
                break

    with open(name_file, "r") as fo:
        lines = fo.readlines()
    heavy_count, hydrogen_count = 0, 0
    with open(out_file, "w") as fo:
        for line in lines:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                symbol = line[76:78].strip()
                if symbol == "H":
                    atom = hydrogen2[Hmaplist[hydrogen_count]]
                    coor = "{:>8.3f}{:>8.3f}{:>8.3f}".format(
                        atom["x"], atom["y"], atom["z"]
                    )
                    fo.write(line[:30] + coor + line[54:])
                    hydrogen_count += 1
                else:
                    atom = heavy2[maplist[heavy_count]]
                    coor = "{:>8.3f}{:>8.3f}{:>8.3f}".format(
                        atom["x"], atom["y"], atom["z"]
                    )
                    fo.write(line[:30] + coor + line[54:])
                    heavy_count += 1
            else:
                fo.write(line)
    logging.warning("DIT can't ensure the results to be correct, CHECK IT by yourself")


def mol_map_call_functions(arguments: list = []):
    """call functions according to arguments"""

    if len(arguments) == 0:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(
        description="mapping coordinates of one pdb file to another pdb file"
    )
    parser.add_argument(
        "-n", "--name", help="pdb file providing infos except coordinates"
    )
    parser.add_argument("-c", "--coor", help="pdb file providing coordinates")
    parser.add_argument("-o", "--output", help="pdb file for output")

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        sys.exit()

    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        sys.exit()

    if len(arguments) == 2:
        logging.error("no parameters, type 'dit <command> -h' for more infos.")
        sys.exit()
    args = parser.parse_args(arguments[2:])
    if method == "mol_map":
        mol_map(args.name, args.coor, args.output)
    else:
        logging.error("unknown method {}".format(method))
        sys.exit()

    logging.info("good day !")


def main():
    mol_map_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
 *******           **                  **********               **        
/**////**         /**          **   **/////**///               /**        
/**    /** **   **/** **    **//** **     /**  ******   ****** /**  ******
/**    /**/**  /**/**/**   /** //***      /** **////** **////**/** **//// 
/**    /**/**  /**/**//** /**   /**       /**/**   /**/**   /**/**//***** 
/**    ** /**  /**/** //****    **        /**/**   /**/**   /**/** /////**
/*******  //******/**  //**    **         /**//****** //****** *** ****** 
///////    ////// //    //    //          //  //////   ////// /// //////  

DuIvyTools (v0.5.0) is a simple analysis and visualization tool for GROMACS result files, written by 杜艾维 (https://github.com/CharlesHahn/DuIvyTools). 

DuIvyTools contains three modules: `Commands` for providing several commands to deal with data and visualization, `FileParser` for providing parser of several type of files (.gro, .mdp, .ndx, .pdb, .xpm, .xvg), `Visualizer` for providing several visualization class, including four plot engine (matplotlib, plotly, gnuplot, and plotext) and several type of plots.

This package is provided to you under GPLv3 License.
"""


"""
DCCM module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to generate dynamical cross correlation matrix from 
covarience dat file generated by `gmx covar`. 

This DCCM module contains:
    - DCCM class

This file is provided to you by GPLv3 license."""


import os
import sys
import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import colors
import logging


class DCCM(object):
    """
    class DCCM are designed to generate dynamical cross correlation matrix
    from `gmx covar` command results.
    """

    def __init__(self) -> None:
        self.resnum = 0
        self.xlabel = ""
        self.ylabel = ""
        self.title = ""

    def read_acsii(self, datfile: str) -> np.ndarray:
        if not os.path.exists(datfile):
            logging.error("no {} in current directory".format(datfile))
            sys.exit()
        if datfile[-4:] != ".dat":
            logging.error("specify a xvg file with suffix dat")
            sys.exit()
        covar = pd.read_csv(datfile, sep=" ", header=None)
        self.resnum = int(np.sqrt((covar.shape[0]) / 3))
        results = pd.DataFrame()
        for i in range(self.resnum):
            one_N = pd.DataFrame()
            for j in range(
                (i * self.resnum) * 3,
                int(len(covar) / self.resnum) * (i + 1),
                self.resnum,
            ):
                df = covar[j : self.resnum + j].reset_index(drop=True)
                one_N = pd.concat([one_N, df], ignore_index=True, axis=1)
            results = pd.concat([results, one_N], ignore_index=True, axis=0)
        results["sum"] = results.sum(axis=1)
        covar = results["sum"].to_numpy()
        covar = covar.reshape(self.resnum, self.resnum)
        return covar

    def covar2corr(self, covar_matrix: np.ndarray) -> np.ndarray:
        corr = np.zeros((self.resnum, self.resnum))
        for i in range(self.resnum):
            for j in range(self.resnum):
                corr[i, j] = covar_matrix[i, j] / np.sqrt(
                    covar_matrix[i, i] * covar_matrix[j, j]
                )
        return corr

    def draw_corr_origin(
        self, corr_matrix: np.ndarray, outpng: str, noshow: bool = False
    ) -> None:
        lis = [
            "#FF80FF",
            "#FFA8FF",
            "#FFD4FF",
            "#FFFCFF",
            "#FFFCFF",
            "#D4FFFF",
            "#A8FFFF",
            "#80FFFF",
        ]
        cm = colors.ListedColormap(lis)
        plt.clf()
        im = plt.imshow(
            corr_matrix, vmin=-1, vmax=1, origin="lower", interpolation="none"
        )
        cb = plt.colorbar(im)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(self.title)
        plt.tight_layout()
        if outpng != None:
            if os.path.exists(outpng):
                logging.error("{} already in current directory".format(outpng))
                sys.exit()
            plt.savefig(outpng, dpi=300)
        if noshow == False:
            plt.show()

    def draw_corr_gaussian(
        self, corr_matrix: np.ndarray, outpng: str, noshow: bool = False
    ) -> None:
        lis = [
            "#FF80FF",
            "#FFA8FF",
            "#FFD4FF",
            "#FFFCFF",
            "#FFFCFF",
            "#D4FFFF",
            "#A8FFFF",
            "#80FFFF",
        ]
        cm = colors.ListedColormap(lis)
        plt.clf()
        im = plt.imshow(
            corr_matrix,
            cmap=cm,
            vmin=-1,
            vmax=1,
            origin="lower",
            interpolation="gaussian",
        )
        cb = plt.colorbar(im)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(self.title)
        plt.tight_layout()
        if outpng != None:
            if os.path.exists(outpng):
                logging.error("{} already in current directory".format(outpng))
                sys.exit()
            plt.savefig(outpng, dpi=300)
        if noshow == False:
            plt.show()

    def draw_corr_bio3d(
        self, corr_matrix: np.ndarray, outpng: str, noshow: bool = False
    ) -> None:
        lis = [
            "#FF80FF",
            "#FFA8FF",
            "#FFD4FF",
            "#FFFCFF",
            "#FFFCFF",
            "#D4FFFF",
            "#A8FFFF",
            "#80FFFF",
        ]
        cm = colors.ListedColormap(lis)
        plt.clf()
        im = plt.contourf(corr_matrix, cmap=cm)
        plt.contour(
            corr_matrix,
            linewidths=0.2,
            levels=[-1.00, -0.75, -0.50, -0.25, 0.25, 0.50, 0.75, 1.00],
        )
        cb = plt.colorbar(im)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(self.title)
        if outpng != None:
            if os.path.exists(outpng):
                logging.error("{} already in current directory".format(outpng))
                sys.exit()
            plt.savefig(outpng, dpi=300)
        if noshow == False:
            plt.show()

    def draw(
        self,
        corr_matrix: np.ndarray,
        mode: str = "bio3d",
        noshow: bool = False,
        outpng: str = None,
    ) -> None:
        if mode == "bio3d":
            self.draw_corr_bio3d(corr_matrix, outpng, noshow)
        elif mode == "gaussian":
            self.draw_corr_gaussian(corr_matrix, outpng, noshow)
        else:
            self.draw_corr_origin(corr_matrix, outpng, noshow)


def dccm_from_ascii(datfile: str, mode, noshow, outpng, xlabel, ylabel, title) -> None:
    dccm = DCCM()
    dccm.xlabel = xlabel
    dccm.ylabel = ylabel
    dccm.title = title
    covar_matrix = dccm.read_acsii(datfile)
    corr_matrix = dccm.covar2corr(covar_matrix)
    dccm.draw(corr_matrix, mode, noshow, outpng)


def dccm_call_functions(arguments: list = None):
    """call functions according to arguments"""

    if arguments == None:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(description="generate mdp file templates")
    parser.add_argument("-f", "--input", help="specify the input file")
    parser.add_argument("-o", "--output", help="file name to output")
    parser.add_argument(
        "-m",
        "--mode",
        default="bio3d",
        help="specify the figure style, 'origin', 'gaussian', 'bio3d'",
    )
    parser.add_argument(
        "-ns", "--noshow", default=False, help="whether not to show figure"
    )
    parser.add_argument(
        "-x", "--xlabel", default="Residue No.", help="specify the xlabel"
    )
    parser.add_argument(
        "-y", "--ylabel", default="Residue No.", help="specify the ylabel"
    )
    parser.add_argument(
        "-t",
        "--title",
        default="Dynamic Cross Correlation Matrix",
        help="specify the title of figure",
    )

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        sys.exit()
    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        sys.exit()

    args = parser.parse_args(arguments[2:])
    if method == "dccm_ascii":
        dccm_from_ascii(
            args.input,
            args.mode,
            args.noshow,
            args.output,
            args.xlabel,
            args.ylabel,
            args.title,
        )
    else:
        logging.error("unknown method {}".format(method))
        sys.exit()

    logging.info("May you good day !")


def main():
    dccm_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
Visualizer_matplotlib module is part of DuIvyTools providing basic visualization tools based on matplotlib.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import time

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors as mplcolors
from matplotlib import patches
from matplotlib.ticker import FormatStrFormatter

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log


class ParentMatplotlib(log):
    """parent class for matplotlib visualizer classes"""

    def __init__(self):
        self.load_style()
        self.figure = plt.figure()

    def load_style(self):
        """load matplotlib style file"""
        style_files = [file for file in os.listdir() if file[-9:] == ".mplstyle"]
        if len(style_files) == 1:
            plt.style.use(style_files[0])
            self.info(f"using matplotlib style sheet from {style_files[0]}")
        elif len(style_files) > 1:
            plt.style.use(style_files[0])
            self.info(
                f"more than one mplstyle files detected, using the {style_files[0]}"
            )
        else:
            data_file_path = os.path.realpath(
                os.path.join(os.getcwd(), os.path.dirname(__file__), "../")
            )
            mplstyle = os.path.join(
                data_file_path, os.path.join("data", "mplstyle", "DIT.mplstyle")
            )
            plt.style.use(mplstyle)
            self.info(
                "using default matplotlib style sheet, to inspect its content, use 'dit show_style'"
            )

    def final(self, outfig: str, noshow: bool) -> None:
        """do final process of drawing figure with matplotlib

        Args:
            outfig (str): the user specified output figure name
            noshow (bool): True for no display the figure
        """
        plt.tight_layout()
        if outfig != None and outfig != "":
            if os.path.exists(outfig):
                time_info = time.strftime("%Y%m%d%H%M%S", time.localtime())
                new_outfig = f'{".".join(outfig.split(".")[:-1])}_{time_info}.{outfig.split(".")[-1]}'
                self.warn(
                    f"{outfig} is already in current directory, save to {new_outfig} for instead."
                )
                outfig = new_outfig
            self.figure.savefig(outfig)
            self.info(f"save figure to {outfig} successfully")
        if noshow == False:
            plt.show()

    def set_xyprecision_xyt_label(self, **kwargs) -> None:
        """set x_precision, y_precision, xlabel, ylabel, title"""
        ax = plt.gca()
        if kwargs["x_precision"] != None:
            x_p = kwargs["x_precision"]
            ax.xaxis.set_major_formatter(FormatStrFormatter(f"%.{x_p}f"))
        if kwargs["y_precision"] != None:
            y_p = kwargs["y_precision"]
            ax.yaxis.set_major_formatter(FormatStrFormatter(f"%.{y_p}f"))
        plt.xlabel(kwargs["xlabel"])
        plt.ylabel(kwargs["ylabel"])
        plt.title(kwargs["title"])

    def set_xytick_precision_xyt_label(self, **kwargs) -> None:
        """set IMAGE x_precision, y_precision, xlabel, ylabel, title"""
        if kwargs["x_precision"] == None:
            kwargs["x_precision"] = 0
        if kwargs["y_precision"] == None:
            kwargs["y_precision"] = 0
        ## set ticks: since matrix, the xtics should all be int, not float
        xrange = len(kwargs["xdata_list"])
        yrange = len(kwargs["ydata_list"])
        xtics, _ = plt.xticks()
        xtics = [int(x) for x in xtics if int(x) < xrange and int(x) >= 0]
        plt.xticks(
            xtics,
            [f'{kwargs["xdata_list"][x]:.{kwargs["x_precision"]}f}' for x in xtics],
        )
        ytics, _ = plt.yticks()
        ytics = [int(y) for y in ytics if int(y) < yrange and int(y) >= 0]
        plt.yticks(
            ytics,
            [f'{kwargs["ydata_list"][y]:.{kwargs["y_precision"]}f}' for y in ytics],
        )

        plt.xlabel(kwargs["xlabel"])
        plt.ylabel(kwargs["ylabel"])
        plt.title(kwargs["title"])


class LineMatplotlib(ParentMatplotlib):
    """A matplotlib line plot class for line plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        highs :List[List[float]]
        lows :List[List[float]]
        alpha :float
        legend_location:str #{inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        for i, data in enumerate(kwargs["data_list"]):
            if len(kwargs["highs"]) != 0 and len(kwargs["lows"]) != 0:
                plt.fill_between(
                    kwargs["xdata_list"][i],
                    kwargs["highs"][i],
                    kwargs["lows"][i],
                    alpha=kwargs["alpha"],
                )
            plt.plot(kwargs["xdata_list"][i], data, label=kwargs["legends"][i])

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            plt.xlim(kwargs["xmin"], kwargs["xmax"])
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            plt.ylim(kwargs["ymin"], kwargs["ymax"])

        if kwargs["legend_location"] == "outside":
            plt.legend(bbox_to_anchor=(1.02, 1.00), loc="upper left")
        else:
            plt.legend()

        self.set_xyprecision_xyt_label(**kwargs)


class ScatterMatplotlib(ParentMatplotlib):
    """A matplotlib scatter plot class for scatter plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        color_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        zmin :float
        zmax :float
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        cmap :str
        colorbar_location:str
        legend_location:str #{inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        shrink = 1.0
        if kwargs["legend_location"] == "outside" and kwargs["colorbar_location"] in [
            None,
            "right",
        ]:
            shrink = 0.5  # avoid overlapping of colorbar and legends
        marker_str = "ov^<>8sp*h.HDdPX"
        for i, data in enumerate(kwargs["data_list"]):
            colors = kwargs["color_list"][i]
            if colors != None:
                plt.scatter(
                    kwargs["xdata_list"][i],
                    data,
                    c=kwargs["color_list"][i],
                    label=kwargs["legends"][i],
                    marker=marker_str[i],
                    cmap=kwargs["cmap"],
                    vmin=kwargs["zmin"],
                    vmax=kwargs["zmax"],
                    alpha=kwargs["alpha"],
                )
                if kwargs["z_precision"] != None:
                    plt.colorbar(
                        label=kwargs["zlabel"],
                        format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                        location=kwargs["colorbar_location"],
                        shrink=shrink,
                    )
                else:
                    plt.colorbar(
                        label=kwargs["zlabel"],
                        location=kwargs["colorbar_location"],
                        shrink=shrink,
                    )
            else:
                plt.scatter(
                    kwargs["xdata_list"][i],
                    data,
                    label=kwargs["legends"][i],
                    marker=marker_str[i],
                    alpha=kwargs["alpha"],
                )

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            plt.xlim(kwargs["xmin"], kwargs["xmax"])
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            plt.ylim(kwargs["ymin"], kwargs["ymax"])

        if kwargs["legend_location"] == "outside":
            plt.legend(bbox_to_anchor=(1.02, 1.00), loc="upper left")
        else:
            plt.legend()

        self.set_xyprecision_xyt_label(**kwargs)


class StackMatplotlib(ParentMatplotlib):
    """A matplotlib stack line plot class for stack line plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        # optional
        highs :List[List[float]]
        lows :List[List[float]]
        alpha :float
        legend_location:str #{inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        for i, _ in enumerate(kwargs["data_list"]):
            plt.fill_between(
                kwargs["xdata_list"][i],
                kwargs["highs"][i],
                kwargs["lows"][i],
                alpha=kwargs["alpha"],
                label=kwargs["legends"][i],
            )

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            plt.xlim(kwargs["xmin"], kwargs["xmax"])
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            plt.ylim(kwargs["ymin"], kwargs["ymax"])

        if kwargs["legend_location"] == "outside":
            plt.legend(bbox_to_anchor=(1.02, 1.00), loc="upper left")
        else:
            plt.legend()

        self.set_xyprecision_xyt_label(**kwargs)


class BarMatplotlib(ParentMatplotlib):
    """A matplotlib bar plot class for bar plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        stds_list :List[List[float]]
        xtitles :List[str]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        legend_location :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        width = 84 // len(kwargs["data_list"]) * 0.01
        x_loc = [x - 0.42 + width / 2.0 for x in range(len(kwargs["data_list"][0]))]
        for i, data in enumerate(kwargs["data_list"]):
            plt.bar(
                [x + width * i for x in x_loc],
                data,
                width,
                yerr=kwargs["stds_list"][i],
                capsize=4,
                label=kwargs["legends"][i],
            )
        plt.xticks(
            [x for x in range(len(kwargs["data_list"][0]))], labels=kwargs["xtitles"]
        )
        plt.axhline(0, color="k", linewidth=1)

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            plt.xlim(kwargs["xmin"], kwargs["xmax"])
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            plt.ylim(kwargs["ymin"], kwargs["ymax"])

        if kwargs["legend_location"] == "outside":
            plt.legend(bbox_to_anchor=(1.02, 1.00), loc="upper left")
        else:
            plt.legend()

        self.set_xyprecision_xyt_label(**kwargs)


class BoxMatplotlib(ParentMatplotlib):
    """A matplotlib box plot class for box plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        color_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        zmin :float
        zmax :float
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        cmap :str
        colorbar_location:str
        mode :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        ## scatter
        loc = 1.0
        if kwargs["mode"] != "withoutScatter":
            loc = 0.75
            for i, data in enumerate(kwargs["data_list"]):
                plt.scatter(
                    np.random.normal(i + 1.25, 0.04, len(data)),
                    data,
                    alpha=kwargs["alpha"],
                    c=kwargs["color_list"][i],
                    cmap=kwargs["cmap"],
                    vmin=kwargs["zmin"],
                    vmax=kwargs["zmax"],
                )
            if kwargs["z_precision"] != None:
                plt.colorbar(
                    label=kwargs["zlabel"],
                    format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                    location=kwargs["colorbar_location"],
                )
            else:
                plt.colorbar(
                    label=kwargs["zlabel"],
                    location=kwargs["colorbar_location"],
                )

        box_positions = [i + loc for i in range(len(kwargs["data_list"]))]
        plt.violinplot(
            kwargs["data_list"],
            showmeans=False,
            showmedians=False,
            showextrema=False,
            positions=box_positions,
        )
        plt.boxplot(
            kwargs["data_list"],
            sym=".",
            meanline=True,
            showmeans=True,
            # patch_artist=True,
            notch=True,
            widths=0.1,
            positions=box_positions,
        )
        plt.xticks([i + 1 for i in range(len(kwargs["data_list"]))], kwargs["legends"])

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            plt.xlim(kwargs["xmin"], kwargs["xmax"])
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            plt.ylim(kwargs["ymin"], kwargs["ymax"])

        self.set_xyprecision_xyt_label(**kwargs)


class RamachandranMatplotlib(ParentMatplotlib):
    """A matplotlib ramachandran plot class for ramachandran plots

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        normals :Dict[Dict[List[float|str]]]
        outliers :Dict[Dict[List[float|str]]]
        rama_pref_values :List[List[float]]
        rama_preferences :Dict[Dict[str:str|List]]
        xlabel :str
        ylabel :str
        title :str
        outfig :str
        noshow :bool
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        ## draw ramachandran plot
        normals = kwargs["normals"]
        outliers = kwargs["outliers"]
        rama_pref_values = kwargs["rama_pref_values"]
        rama_preferences = kwargs["rama_preferences"]
        for key in ["General", "GLY", "Pre-PRO", "PRO"]:
            if len(normals[key]["phi"]) + len(outliers[key]["phi"]) == 0:
                continue
            plt.clf()
            plt.imshow(
                rama_pref_values[key],
                cmap=mplcolors.ListedColormap(rama_preferences[key]["cmap"]),
                norm=mplcolors.BoundaryNorm(
                    rama_preferences[key]["bounds"],
                    mplcolors.ListedColormap(rama_preferences[key]["cmap"]).N,
                ),
                extent=(-180, 180, 180, -180),
            )
            plt.scatter(normals[key]["phi"], normals[key]["psi"], s=8)
            plt.scatter(outliers[key]["phi"], outliers[key]["psi"], s=8)
            plt.xlim([-180, 180])
            plt.ylim([-180, 180])
            plt.xticks([-180, -120, -60, 0, 60, 120, 180])
            plt.yticks([-180, -120, -60, 0, 60, 120, 180])
            plt.tick_params(left=False, bottom=False, top=False, right=False)
            if kwargs["title"] == None:
                title = key
            else:
                title = kwargs["title"]

            ax = plt.gca()
            if kwargs["x_precision"] != None:
                x_p = kwargs["x_precision"]
                ax.xaxis.set_major_formatter(FormatStrFormatter(f"%.{x_p}f"))
            if kwargs["y_precision"] != None:
                y_p = kwargs["y_precision"]
                ax.yaxis.set_major_formatter(FormatStrFormatter(f"%.{y_p}f"))
            plt.xlabel(kwargs["xlabel"])
            plt.ylabel(kwargs["ylabel"])
            plt.title(title)

            plt.tight_layout()
            outfig = kwargs["outfig"]
            noshow = kwargs["noshow"]
            if outfig != None:
                if os.path.exists(outfig):
                    time_info = time.strftime("%Y%m%d%H%M%S", time.localtime())
                    new_outfig = f'{".".join(outfig.split(".")[:-1])}_{time_info}.{outfig.split(".")[-1]}'
                    self.warn(
                        f"{outfig} is already in current directory, save to {new_outfig} for instead."
                    )
                    outfig = new_outfig
                outfig = outfig.split(".")[0] + "_" + key + ".png"
                self.figure.savefig(outfig)
                self.info(f"save figure to {outfig} successfully")
            if noshow == False:
                plt.show()


class ImshowMatplotlib(ParentMatplotlib):
    """A matplotlib imshow plot class for heatmap

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[float]
        ydata_list :List[float]
        legends :List[str]
        color_list :List[str]
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        zmin :float
        zmax :float
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        legend_location :str
        colorbar_location :str
        interpolation :str
        cmap :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        if kwargs["fig_type"] != "Continuous":
            color_map = mplcolors.ListedColormap(kwargs["color_list"])
            im = plt.imshow(
                kwargs["data_list"],
                alpha=kwargs["alpha"],
                cmap=color_map,
                origin="lower",
            )
            legend_patches = []
            for ind, note in enumerate(kwargs["legends"]):
                leg_patch = patches.Patch(color=kwargs["color_list"][ind], label=note)
                legend_patches.append(leg_patch)
            if kwargs["legend_location"] == "outside":
                plt.legend(
                    handles=legend_patches,
                    bbox_to_anchor=(1.02, 1.00),
                    loc="upper left",
                    borderaxespad=0,
                )
            else:
                plt.legend(handles=legend_patches)
        else:
            im = plt.imshow(
                kwargs["data_list"],
                interpolation=kwargs["interpolation"],
                alpha=kwargs["alpha"],
                cmap=kwargs["cmap"],
                origin="lower",
                vmin=kwargs["zmin"],
                vmax=kwargs["zmax"],
            )
            if kwargs["z_precision"] != None:
                plt.colorbar(
                    im,
                    label=kwargs["zlabel"],
                    format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                    location=kwargs["colorbar_location"],
                )
            else:
                plt.colorbar(
                    im, label=kwargs["zlabel"], location=kwargs["colorbar_location"]
                )

        self.set_xytick_precision_xyt_label(**kwargs)


class PcolormeshMatplotlib(ParentMatplotlib):
    """A matplotlib pcolormesh plot class for heatmap

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[float]
        ydata_list :List[float]
        legends :List[str]
        color_list :List[str]
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        zmin :float
        zmax :float
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        legend_location :str
        colorbar_location :str
        cmap :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        if len(kwargs["data_list"]) <= 1 or len(kwargs["data_list"][0]) <= 1:
            self.error("!!! pcolormesh unable to proper deal with 1 dimension data !!!")
        # to avoid repeat series in X and Y data
        xdata_index = [i for i in range(len(kwargs["xdata_list"]))]
        ydata_index = [i for i in range(len(kwargs["ydata_list"]))]
        xdata_index, ydata_index = np.meshgrid(xdata_index, ydata_index)
        if kwargs["fig_type"] != "Continuous":
            self.info(
                f"""using user-defined colors, the original colors in xpm are: {kwargs["color_list"]}"""
            )
            colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
            if len(colors) < len(kwargs["legends"]):
                self.error(
                    f"""you need to specify at least {len(kwargs["legends"])} colors to show discrete xpm figure"""
                )
            colors = colors[: len(kwargs["legends"])]
            color_map = mplcolors.ListedColormap(colors)
            im = plt.pcolormesh(
                xdata_index,
                ydata_index,
                kwargs["data_list"],
                alpha=kwargs["alpha"],
                cmap=color_map,
                shading="auto",
            )
            legend_patches = []
            for ind, note in enumerate(kwargs["legends"]):
                leg_patch = patches.Patch(color=colors[ind], label=note)
                legend_patches.append(leg_patch)
            if kwargs["legend_location"] == "outside":
                plt.legend(
                    handles=legend_patches,
                    bbox_to_anchor=(1.02, 1.00),
                    loc="upper left",
                    borderaxespad=0,
                )
            else:
                plt.legend(handles=legend_patches)
        else:
            im = plt.pcolormesh(
                xdata_index,
                ydata_index,
                kwargs["data_list"],
                alpha=kwargs["alpha"],
                cmap=kwargs["cmap"],
                shading="auto",
                vmin=kwargs["zmin"],
                vmax=kwargs["zmax"],
            )
            if kwargs["z_precision"] != None:
                plt.colorbar(
                    im,
                    label=kwargs["zlabel"],
                    format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                    location=kwargs["colorbar_location"],
                )
            else:
                plt.colorbar(
                    im, label=kwargs["zlabel"], location=kwargs["colorbar_location"]
                )

        self.set_xytick_precision_xyt_label(**kwargs)


class ThreeDimensionMatplotlib(ParentMatplotlib):
    """A matplotlib 3d plot class for heatmap

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[float]
        ydata_list :List[float]
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        colorbar_location :str
        cmap :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        if len(kwargs["data_list"]) <= 1 or len(kwargs["data_list"][0]) <= 1:
            self.error("!!! 3D plot unable to proper deal with 1 dimension data !!!")
        if kwargs["cmap"] == None:
            if plt.rcParams.get("image.cmap", None) == None:
                self.warn(
                    "you have not set the colormap through commands or image.cmap of mplstyle file, the color of 3D plot may not be pretty"
                )
            kwargs["cmap"] = plt.rcParams["image.cmap"]

        # to avoid repeat series in X and Y data
        xdata_index = [i for i in range(len(kwargs["xdata_list"]))]
        ydata_index = [i for i in range(len(kwargs["ydata_list"]))]
        xdata_index, ydata_index = np.meshgrid(xdata_index, ydata_index)
        ax = self.figure.add_subplot(projection="3d")
        im = ax.plot_surface(
            xdata_index,
            ydata_index,
            kwargs["data_list"],
            alpha=kwargs["alpha"],
            cmap=kwargs["cmap"],
            linewidth=0,
            antialiased=False,
        )
        # set the 2d surface location
        offset = (
            np.floor(np.min(kwargs["data_list"]))
            - np.floor(np.max(kwargs["data_list"]) - np.min(kwargs["data_list"])) / 30
        )
        ax.contourf(
            xdata_index,
            ydata_index,
            kwargs["data_list"],
            cmap=kwargs["cmap"],
            zdir="z",
            offset=offset,
            vmin=kwargs["zmin"],
            vmax=kwargs["zmax"],
        )
        if kwargs["z_precision"] != None:
            plt.colorbar(
                im,
                label=kwargs["zlabel"],
                format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                location=kwargs["colorbar_location"],
            )
        else:
            plt.colorbar(
                im, label=kwargs["zlabel"], location=kwargs["colorbar_location"]
            )

        ax.set_zlabel(kwargs["zlabel"])
        self.set_xytick_precision_xyt_label(**kwargs)


class ContourMatplotlib(ParentMatplotlib):
    """A matplotlib contour plot class for heatmap

    Args:
        ParentMatplotlib (object): matplotlib parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[float]
        ydata_list :List[float]
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        colorbar_location :str
        cmap :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        if len(kwargs["data_list"]) <= 1 or len(kwargs["data_list"][0]) <= 1:
            self.error("!!! contour unable to proper deal with 1 dimension data !!!")
        # to avoid repeat series in X and Y data
        xdata_index = [i for i in range(len(kwargs["xdata_list"]))]
        ydata_index = [i for i in range(len(kwargs["ydata_list"]))]
        xdata_index, ydata_index = np.meshgrid(xdata_index, ydata_index)
        plt.contourf(
            xdata_index,
            ydata_index,
            kwargs["data_list"],
            cmap=kwargs["cmap"],
            vmin=kwargs["zmin"],
            vmax=kwargs["zmax"],
        )
        if kwargs["z_precision"] != None:
            plt.colorbar(
                label=kwargs["zlabel"],
                format=FormatStrFormatter(f"""%.{kwargs["z_precision"]}f"""),
                location=kwargs["colorbar_location"],
            )
        else:
            plt.colorbar(label=kwargs["zlabel"], location=kwargs["colorbar_location"])

        self.set_xytick_precision_xyt_label(**kwargs)



"""
 *******           **                  **********               **        
/**////**         /**          **   **/////**///               /**        
/**    /** **   **/** **    **//** **     /**  ******   ****** /**  ******
/**    /**/**  /**/**/**   /** //***      /** **////** **////**/** **//// 
/**    /**/**  /**/**//** /**   /**       /**/**   /**/**   /**/**//***** 
/**    ** /**  /**/** //****    **        /**/**   /**/**   /**/** /////**
/*******  //******/**  //**    **         /**//****** //****** *** ****** 
///////    ////// //    //    //          //  //////   ////// /// //////  

DuIvyTools is a simple analysis and visualization tool for GROMACS result files
written by 杜艾维 (https://github.com/CharlesHahn/DuIvyTools). 

DuIvyTools package contains two DuIvyTools actually now. One is `DuIvyTools` which is the newest DuIvyTools v0.5 and noted as `dit`, the other is the `DuIvyTools_old` which is v0.4.8 and noted as `dito`.
Since reconstruction were done for v0.5.0 and some features were dropped, the v0.4.8 were kept in package to keep integrity temporarily. 

This package is provided to you under GPLv3 License.
"""